from typing import Optional, List

import psycopg2
from psycopg2 import extensions

from nehushtan.postgres.PgSQLKitConfig import PgSQLKitConfig
from nehushtan.postgres.PgSQLResultSet import PgSQLResultSet


class PgSQLKit:
    """
    Since 0.4.22
    """

    def __init__(self, config: PgSQLKitConfig):
        self.__connection = psycopg2.connect(config.dsn())
        self.set_auto_commit(config.autocommit())

    def get_raw_connection(self):
        """
        use `with` this result to handle a transcation in autocommit mode.
        """
        return self.__connection

    def get_dsn_of_connection(self) -> str:
        """
        Read-only string containing the connection string used by the connection.
        If a password was specified in the connection string it will be obscured.
        """
        return self.__connection.dsn

    def get_cursor(self, scrollable: bool = None, withhold: bool = None):
        cursor = self.__connection.cursor()
        if scrollable is not None:
            cursor.scrollable = scrollable
        if withhold is not None:
            cursor.withhold = withhold
        return cursor

    def close_connection(self):
        """
        Close the connection now (rather than whenever del is executed).
            The connection will be unusable from this point forward;
                an InterfaceError will be raised if any operation is attempted with the connection.
            The same applies to all cursor objects trying to use the connection.
            Note that closing a connection without committing the changes first
                will cause any pending change to be discarded as if a ROLLBACK was performed
                    (unless a different isolation level has been selected: see set_isolation_level()).

        Changed in version 2.2:
            previously an explicit ROLLBACK was issued by Psycopg on close().
            The command could have been sent to the backend at an inappropriate time,
            so Psycopg currently relies on the backend to implicitly discard uncommitted changes.
            Some middleware are known to behave incorrectly though when the connection is closed during a transaction
                (when status is STATUS_IN_TRANSACTION),
                e.g. PgBouncer reports an unclean server and discards the connection.
            To avoid this problem you can ensure to terminate the transaction with a commit()/rollback() before closing.
        """
        if self.__connection:
            self.__connection.close()

    def is_closed(self) -> bool:
        """
        Read-only integer attribute: 0 if the connection is open, nonzero if it is closed or broken.
        """
        return True if self.__connection.closed != 0 else False

    def __del__(self):
        self.close_connection()

    def execute(self, sql: str, variables=None):
        cursor = self.get_cursor()
        """
        cursor.execute(query, vars=None)
        
        Execute a database operation (query or command).
        Parameters may be provided as sequence or mapping and will be bound to variables in the operation. 
            Variables are specified either with positional (%s) or named (%(name)s) placeholders. 
            See Passing parameters to SQL queries.
        The method returns None. 
            If a query was executed, the returned values can be retrieved using fetch*() methods.
        """
        if variables:
            cursor.execute(sql, variables)
        else:
            cursor.execute(sql)
        return PgSQLResultSet(cursor)

    def mogrify(self, sql: str, variables=None) -> bytes:
        cursor = self.get_cursor()
        """
        mogrify(operation[, parameters])

        Return a query string after arguments binding. 
            The string returned is exactly the one 
                that would be sent to the database running the execute() method or similar.
        The returned string is always a bytes string.
        """
        if variables:
            return cursor.mogrify(sql, variables)
        else:
            return cursor.mogrify(sql)

    def call_function(self, fucntion_name: str, parameters: None):
        cursor = self.get_cursor()
        """
        callproc(procname[, parameters])

        Call a stored database procedure with the given name. 
            The sequence of parameters must contain one entry for each argument that the procedure expects. 
            Overloaded procedures are supported. 
            Named parameters can be used by supplying the parameters as a dictionary.
        This function is, at present, not DBAPI-compliant. 
            The return value is supposed to consist of the sequence of parameters 
                with modified output and input or output parameters. 
            In future versions, the DBAPI-compliant return value may be implemented, 
                but for now the function returns None.
        The procedure may provide a result set as output. 
            This is then made available through the standard fetch*() methods.

        Changed in version 2.7: added support for named arguments.
        Note 
            callproc() can only be used with PostgreSQL functions, 
                not with the procedures introduced in PostgreSQL 11, 
                which require the CALL statement to run. 
            Please use a normal execute() to run them.
        """
        if parameters:
            cursor.callproc(fucntion_name, parameters)
        else:
            cursor.callproc(fucntion_name)
        return PgSQLResultSet(cursor)

    def cancel(self):
        """
        Cancel the current database operation.
        The method interrupts the processing of the current operation.
            If no query is being executed, it does nothing.
            You can call this function from a different thread than the one currently executing a database operation,
            for instance if you want to cancel a long running query if a button is pushed in the UI.
            Interrupting query execution will cause the cancelled method to raise a QueryCanceledError.
            Note that the termination of the query is not guaranteed to succeed: see the documentation for PQcancel().
        New in version 2.3.
        """
        self.__connection.cancel()

    def reset(self):
        """
        Reset the connection to the default.
        The method rolls back an eventual pending transaction
            and executes the PostgreSQL RESET and SET SESSION AUTHORIZATION to revert the session to the default values.
        A two-phase commit transaction prepared using tpc_prepare() will remain in the database available for recover.
        New in version 2.0.12.
        """
        self.__connection.reset()

    def commit(self):
        """
        Commit any pending transaction to the database.
        By default, Psycopg opens a transaction before executing the first command:
            if commit() is not called, the effect of any data manipulation will be lost.
        The connection can be also set in “autocommit” mode:
            no transaction is automatically open, commands have immediate effect.
            See [Transactions control](https://www.psycopg.org/docs/usage.html#transactions-control) for details.
        Changed in version 2.5:
            if the connection is used in a with statement,
                the method is automatically called if no exception is raised in the with block.
        """
        self.__connection.commit()

    def rollback(self):
        """
        Roll back to the start of any pending transaction.
            Closing a connection without committing the changes first will cause an implicit rollback to be performed.
        Changed in version 2.5:
            if the connection is used in a with statement,
            the method is automatically called if an exception is raised in the with block.
        """
        return self.__connection.rollback()

    def __xid(self, format_id, gtrid, bqual):
        """
        Returns a Xid instance to be passed to the tpc_*() methods of this connection.
            The argument types and constraints are explained in Two-Phase Commit protocol support.
        The values passed to the method will be available on the returned object as the members format_id, gtrid, bqual.
        The object also allows accessing to these members and unpacking as a 3-items tuple.
        """
        return self.__connection.xid(format_id, gtrid, bqual)

    def __tpc_begin(self, xid):
        """
        Begins a TPC transaction with the given transaction ID xid.
        This method should be called outside of a transaction
            (i.e. nothing may have executed since the last commit() or rollback()
                and connection.status is STATUS_READY).
        Furthermore, it is an error to call commit() or rollback() within the TPC transaction:
            in this case a ProgrammingError is raised.
        The xid may be either an object returned by the xid() method or a plain string:
            the latter allows to create a transaction using the provided string as PostgreSQL transaction id.
            See also tpc_recover().
        """
        return self.__connection.tpc_begin(xid)

    def __tpc_prepare(self):
        """
        Performs the first phase of a transaction started with tpc_begin().
            A ProgrammingError is raised if this method is used outside of a TPC transaction.
        After calling tpc_prepare(), no statements can be executed until tpc_commit() or tpc_rollback() will be called.
        The reset() method can be used to restore the status of the connection to STATUS_READY:
            the transaction will remain prepared in the database
                and will be possible to finish it with tpc_commit(xid) and tpc_rollback(xid).
        See also the PREPARE TRANSACTION PostgreSQL command.
        """
        return self.__connection.tpc_prepare()

    def __tpc_commit(self, xid=None):
        """
        When called with no arguments, tpc_commit() commits a TPC transaction previously prepared with tpc_prepare().
        If tpc_commit() is called prior to tpc_prepare(), a single phase commit is performed.
        A transaction manager may choose to do this
            if only a single resource is participating in the global transaction.
        When called with a transaction ID xid, the database commits the given transaction.
        If an invalid transaction ID is provided, a ProgrammingError will be raised.
        This form should be called outside of a transaction, and is intended for use in recovery.
        On return, the TPC transaction is ended.
        See also the COMMIT PREPARED PostgreSQL command.
        """
        if xid:
            return self.__connection.tpc_commit(xid)
        else:
            return self.__connection.tpc_commit()

    def __tpc_rollback(self, xid=None):
        """
        When called with no arguments, tpc_rollback() rolls back a TPC transaction.
        It may be called before or after tpc_prepare().
        When called with a transaction ID xid, it rolls back the given transaction.
        If an invalid transaction ID is provided, a ProgrammingError is raised.
        This form should be called outside of a transaction, and is intended for use in recovery.
        On return, the TPC transaction is ended.
        See also the ROLLBACK PREPARED PostgreSQL command.
        """
        if xid:
            return self.__connection.tpc_rollback(xid)
        else:
            return self.__connection.tpc_rollback()

    def __tpc_recover(self):
        """
        Returns a list of Xid representing pending transactions, suitable for use with tpc_commit() or tpc_rollback().
        If a transaction was not initiated by Psycopg, the returned Xids will have attributes
            format_id and bqual set to None
            and the gtrid set to the PostgreSQL transaction ID:
                such Xids are still usable for recovery.
        Psycopg uses the same algorithm of the PostgreSQL JDBC driver to encode a XA triple in a string,
            so transactions initiated by a program using such driver should be unpacked correctly.
        Xids returned by tpc_recover() also have extra attributes prepared,
            owner, database populated with the values read from the server.
        See also the pg_prepared_xacts system view.
        """
        return self.__connection.tpc_recover()

    def set_session(self, isolation_level=None, readonly=None, deferrable=None, autocommit=None):
        """
        Set one or more parameters for the next transactions or statements in the current session.
        The function must be invoked with no transaction in progress.

        Parameters:
        * isolation_level
            set the isolation level for the next transactions/statements.
            The value can be one of the literal values
                READ UNCOMMITTED,
                READ COMMITTED,
                REPEATABLE READ,
                SERIALIZABLE
                or the equivalent constant defined in the extensions module.
        * readonly
            if True, set the connection to read only;
            read/write if False.
        * deferrable
            if True, set the connection to deferrable;
            non deferrable if False.
            Only available from PostgreSQL 9.1.
        * autocommit
            switch the connection to autocommit mode:
                not a PostgreSQL session setting but an alias for setting the autocommit attribute.
        """
        modification = {}
        if isolation_level:
            modification['isolation_level'] = isolation_level
        if readonly:
            modification['readonly'] = readonly
        if deferrable:
            modification['deferrable'] = deferrable
        if autocommit:
            modification['autocommit'] = autocommit
        self.set_session(**modification)

    def is_auto_commit(self) -> bool:
        return self.__connection.autocommit

    def set_auto_commit(self, auto_commit: bool):
        self.__connection.autocommit = auto_commit

    def get_isolation_level(self) -> Optional[int]:
        """
        use psycopg2.extensions.ISOLATION_LEVEL_*
        """
        return self.__connection.isolation_level

    def set_isolation_level(self, isolation_level: Optional[int]):
        self.__connection.isolation_level = isolation_level

    def is_readonly(self) -> Optional[bool]:
        return self.__connection.readonly

    def set_readonly(self, readonly: Optional[bool]):
        self.__connection.readonly = readonly

    def is_deferrable(self) -> Optional[bool]:
        return self.__connection.deferrable

    def set_deferrable(self, deferrable: Optional[bool]):
        self.__connection.deferrable = deferrable

    def get_encoding(self):
        """
        Read the client encoding for the current session.
        The default is the encoding defined by the database.
        It should be one of the characters set supported by PostgreSQL
        """
        return self.__connection.encoding

    def set_client_encoding(self, enc):
        """
        set the client encoding for the current session.
        The default is the encoding defined by the database.
        It should be one of the characters set supported by PostgreSQL
        """
        return self.__connection.set_client_encoding(enc)

    def get_notices(self) -> List[str]:
        """
        A list containing all the database messages sent to the client during the session.
        """
        return self.__connection.notices

    def get_notifies(self) -> list:
        """
        List of Notify objects containing asynchronous notifications received by the session.
        For other details see Asynchronous notifications.
        """
        return self.__connection.notifies

    def get_conection_info(self):
        """
        A ConnectionInfo object exposing information about the native libpq connection.
        """
        return self.__connection.info

    def get_status(self):
        """
        A read-only integer representing the status of the connection.
            Symbolic constants for the values are defined in the module psycopg2.extensions:
                see Connection status constants for the available values.
        The status is undefined for closed connections.
        """
        return self.__connection.status

    def __lobject(self, oid=0, mode=0, new_oid=0, new_file=None, lobject_factory=extensions.lobject):
        return self.__connection.lobject(oid, mode, new_oid, new_file, lobject_factory)

    def get_transaction_status(self):
        """
        Also available as info.transaction_status.
        Return the current session transaction status as an integer.
            Symbolic constants for the values are defined in the module psycopg2.extensions:
                see Transaction status constants for the available values.
        See also libpq docs for PQtransactionStatus() for details.
        """
        return self.__connection.get_transaction_status()

    def get_protocol_version(self):
        """
        Also available as info.protocol_version.
        A read-only integer representing frontend/backend protocol being used.
            Currently Psycopg supports only protocol 3,
                which allows connection to PostgreSQL server from version 7.4.
            Psycopg versions previous than 2.3 support both protocols 2 and 3.
        See also libpq docs for PQprotocolVersion() for details.
        New in version 2.0.12.
        """
        return self.__connection.protocol_version

    def get_server_version(self):
        """
        Also available as info.server_version.
        A read-only integer representing the backend version.
        The number is formed by converting the major, minor, and revision numbers
            into two-decimal-digit numbers and appending them together.
        For example, version 8.1.5 will be returned as 80105.
        See also libpq docs for PQserverVersion() for details.
        New in version 2.0.12.
        """
        return self.__connection.server_version

    def get_backend_pid(self):
        """
        Also available as info.backend_pid.
        Returns the process ID (PID) of the backend server process you connected to.
            Note that if you use a connection pool service such as PgBouncer this value will not be updated
                if your connection is switched to a different backend.
        Note that the PID belongs to a process executing on the database server host, not the local host!
        See also libpq docs for PQbackendPID() for details.
        New in version 2.0.8.
        """
        return self.__connection.get_backend_pid()

    def get_parameter_status(self, parameter):
        """
        Also available as info.parameter_status().
        Look up a current parameter setting of the server.
        Potential values for parameter are:
            server_version, server_encoding, client_encoding, is_superuser, session_authorization,
            DateStyle, TimeZone, integer_datetimes, and standard_conforming_strings.
        If server did not report requested parameter, return None.
        See also libpq docs for PQparameterStatus() for details.
        New in version 2.0.12.
        """
        return self.__connection.get_parameter_status(parameter)

    def get_dsn_parameters(self):
        """
        Also available as info.dsn_parameters.
        Get the effective dsn parameters for the connection as a dictionary.
        The password parameter is removed from the result.
        Requires libpq >= 9.3.
        See also libpq docs for PQconninfo() for details.
        New in version 2.7.
        """
        return self.__connection.get_dsn_parameters()



from typing import Union, List


class CSISequence:
    """
    Since 0.4.13
    """

    def __init__(self, parameter='', intermediate='', final=''):
        """
        CSI序列由ESC [、若干个（包括0个）“参数字节”、若干个“中间字节”，以及一个“最终字节”组成。

        参数字节	0x30–0x3F	0–9:;<=>?
        中间字节	0x20–0x2F	空格、!"#$%&'()*+,-./
        最终字节	0x40–0x7E	@A–Z[\]^_`a–z{|}~

        """

        self.parameter = parameter
        self.intermediate = intermediate
        self.final = final

    def to_string(self):
        return '\033[' + self.parameter + self.intermediate + self.final

    def __str__(self):
        return self.to_string()

    # 光标向指定的方向移动{\displaystyle n}n（默认1）格。如果光标已在屏幕边缘，则无效。

    @staticmethod
    def csi_for_cursor_up(n: int = 1):
        return CSISequence(f'{n}', '', 'A')

    @staticmethod
    def csi_for_cursor_down(n: int = 1):
        return CSISequence(f'{n}', '', 'B')

    @staticmethod
    def csi_for_cursor_forward(n: int = 1):
        return CSISequence(f'{n}', '', 'C')

    @staticmethod
    def csi_for_cursor_back(n: int = 1):
        return CSISequence(f'{n}', '', 'D')

    @staticmethod
    def csi_for_cursor_next_line(n: int = 1):
        """
        光标移动到下面第 n（默认1）行的开头。（非ANSI.SYS）
        """
        return CSISequence(f'{n}', '', 'E')

    @staticmethod
    def csi_for_cursor_previous_line(n: int = 1):
        """
        光标移动到上面第 n（默认1）行的开头。（非ANSI.SYS）
        """
        return CSISequence(f'{n}', '', 'F')

    @staticmethod
    def csi_for_cursor_horizontal_absolute(n: int = 1):
        """
        光标移动到第 n（默认1）列。（非ANSI.SYS）
        """
        return CSISequence(f'{n}', '', 'G')

    @staticmethod
    def csi_for_cursor_move_to_position(row_number: int = 1, column_number: int = 1):
        """
        光标移动到第 row_number 行、第 column_number 列。
        值从1开始，且默认为1（左上角）。
        例如CSI ;5H和CSI 1;5H含义相同；
        CSI 17;H、CSI 17H和CSI 17;1H三者含义相同。
        """
        return CSISequence(f'{row_number};{column_number}', '', 'H')

    @staticmethod
    def csi_for_cursor_format_to_position(row_number: int = 1, column_number: int = 1):
        """
        Horizontal Vertical Position
        Same as CUP, but counts as a format effector function (like CR or LF)
            rather than an editor function (like CUD or CNL).
        This can lead to different handling in certain terminal modes.
        """
        return CSISequence(f'{row_number};{column_number}', '', 'f')

    @staticmethod
    def csi_for_erase_area(n: int = 0):
        """
        清除屏幕的部分区域。
        如果n是0（或缺失），则清除从光标位置到屏幕末尾的部分。
        如果n是1，则清除从光标位置到屏幕开头的部分。
        如果n是2，则清除整个屏幕（在DOS ANSI.SYS中，光标还会向左上方移动）。
        如果n是3，则清除整个屏幕，并删除回滚缓存区中的所有行（这个特性是xterm添加的，其他终端应用程序也支持）。
        """
        return CSISequence(f'{n}', '', 'J')

    @staticmethod
    def csi_for_erase_line(n: int = 0):
        """
        清除行内的部分区域。
        如果n是0（或缺失），清除从光标位置到该行末尾的部分。
        如果n是1，清除从光标位置到该行开头的部分。
        如果n是2，清除整行。光标位置不变。
        """
        return CSISequence(f'{n}', '', 'K')

    @staticmethod
    def csi_for_scroll_up(n: int = 1):
        """
        整页向上滚动n（默认1）行。新行添加到底部。（非ANSI.SYS）
        """
        return CSISequence(f'{n}', '', 'S')

    @staticmethod
    def csi_for_scroll_down(n: int = 1):
        """
        整页向下滚动n（默认1）行。新行添加到顶部。（非ANSI.SYS）
        """
        return CSISequence(f'{n}', '', 'T')

    @staticmethod
    def csi_for_select_graphic_rendition(sgr: Union[str, int, List[Union[str, int]]] = None):
        """
        选择图形再现（Select Graphic Rendition）
        """
        if type(sgr) is str:
            parameters = [sgr]
        elif type(sgr) is int:
            parameters = [f'{sgr}']
        elif type(sgr) is list:
            parameters = []
            for x in sgr:
                if type(x) is str:
                    parameters.append(x)
                elif type(x) is int:
                    parameters.append(f'{x}')
                else:
                    raise SyntaxError('SGR FORMAT ERROR')
        else:
            raise SyntaxError('SGR FORMAT ERROR')

        return CSISequence(';'.join(parameters), '', 'm')

    @staticmethod
    def csi_for_device_status_report():
        """
        Reports the cursor position (CPR) by transmitting `ESC[n;mR`, where n is the row and m is the column.)
        """
        return CSISequence('6', '', 'n')



import json
from typing import List

from nehushtan.score.ParsedLine import ParsedLine
from nehushtan.score.ScoreUnit import ScoreUnit


class ParsedLineAsScore(ParsedLine):
    def __init__(self, score_units: List[ScoreUnit]):
        super().__init__()
        self.__score_units = score_units

    def get_score_units(self):
        return self.__score_units

    def get_score_unit(self, index: int):
        return self.__score_units[index]

    def __str__(self):
        y = []
        for su in self.__score_units:
            x = su.__repr__()
            y.append(x)
        return "ParsedLineAsScore: " + json.dumps(y)



#  Copyright (c) 2020. Sinri Edogawa
from typing import Iterable, Union, Optional

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLKit import MySQLKit


class MySQLCondition:

    def __init__(self, field: str, operate: str, value: any, addition: any = None,
                 field_qualifier: Optional[str] = None):
        self._field_qualifier = field_qualifier
        self._field = field
        self._operate = operate
        self._value = value

        if self._operate == constant.MYSQL_CONDITION_OP_LIKE or self._operate == constant.MYSQL_CONDITION_OP_NOT_LIKE:
            self._value = MySQLKit.quote_offline(self._value)
            if addition == constant.MYSQL_CONDITION_LIKE_LEFT_WILDCARD:
                self._value = f"concat('%',{self._value})"
            elif addition == constant.MYSQL_CONDITION_LIKE_RIGHT_WILDCARD:
                self._value = f"concat({self._value},'%')"
            elif addition == constant.MYSQL_CONDITION_LIKE_BOTH_WILDCARD:
                self._value = f"concat('%',{self._value},'%')"
            # else treat as EQUAL of string

    def get_field_expression(self) -> str:
        e = f'`{self._field}`'
        if self._field_qualifier is not None:
            e = f'`{self._field_qualifier}`.{e}'
        return e

    @staticmethod
    def make_equal(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_EQ, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_equal_null_safe(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NULL_SAFE_EQUAL, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_not_equal(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NEQ, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_in_array(field: str, array: Union[tuple, list], field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_IN, value=array,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_not_in_array(field: str, array: Union[tuple, list], field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NOT_IN, value=array,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_greater_than(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_GT, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_equal_or_greater_than(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_EGT, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_less_than(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_LT, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_equal_or_less_than(field: str, value: any, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_ELT, value=value,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_is_null(field: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_IS,
                              value=constant.MYSQL_CONDITION_CONST_NULL, field_qualifier=field_qualifier)

    @staticmethod
    def make_is_not_null(field: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_IS_NOT,
                              value=constant.MYSQL_CONDITION_CONST_NULL, field_qualifier=field_qualifier)

    @staticmethod
    def make_between(field: str, left, right, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_BETWEEN, value=(left, right),
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_not_between(field: str, left, right, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NOT_BETWEEN, value=(left, right),
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_string_with_prefix(field: str, prefix: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_LIKE, value=prefix,
                              addition=constant.MYSQL_CONDITION_LIKE_RIGHT_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_with_suffix(field: str, suffix: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_LIKE, value=suffix,
                              addition=constant.MYSQL_CONDITION_LIKE_LEFT_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_contains(field: str, sub_string: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_LIKE, value=sub_string,
                              addition=constant.MYSQL_CONDITION_LIKE_BOTH_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_without_prefix(field: str, prefix: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NOT_LIKE, value=prefix,
                              addition=constant.MYSQL_CONDITION_LIKE_RIGHT_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_without_suffix(field: str, suffix: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NOT_LIKE, value=suffix,
                              addition=constant.MYSQL_CONDITION_LIKE_LEFT_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_contains_not(field: str, sub_string: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_OP_NOT_LIKE, value=sub_string,
                              addition=constant.MYSQL_CONDITION_LIKE_BOTH_WILDCARD, field_qualifier=field_qualifier)

    @staticmethod
    def make_string_is_null_or_empty(field: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_MACRO_IS_NULL_OR_EMPTY_STRING, value=None,
                              field_qualifier=field_qualifier)

    @staticmethod
    def make_string_is_not_null_nor_empty(field: str, field_qualifier: Optional[str] = None):
        return MySQLCondition(field=field, operate=constant.MYSQL_CONDITION_MACRO_IS_NOT_NULL_NOR_EMPTY_STRING,
                              value=None, field_qualifier=field_qualifier)

    @staticmethod
    def make_and(conditions: Iterable["MySQLCondition"]):
        return MySQLCondition.make_intersection_of_conditions(conditions)

    @staticmethod
    def make_intersection_of_conditions(conditions: Iterable["MySQLCondition"]):
        return MySQLCondition('', constant.MYSQL_CONDITION_OP_PARENTHESES_AND, conditions)

    @staticmethod
    def make_or(conditions: Iterable["MySQLCondition"]):
        return MySQLCondition.make_union_of_conditions(conditions)

    @staticmethod
    def make_union_of_conditions(conditions: Iterable["MySQLCondition"]):
        return MySQLCondition('', constant.MYSQL_CONDITION_OP_PARENTHESES_OR, conditions)

    @staticmethod
    def make_raw_expression(condition_sql: str):
        return MySQLCondition('', constant.MYSQL_CONDITION_MACRO_RAW_EXPRESSION, condition_sql)

    def __str__(self):
        return self.organize_to_sql()

    def organize_to_sql(self) -> str:
        """
        Since 0.1.4
        :return:
        """
        if (
                constant.MYSQL_CONDITION_OP_EQ,
                constant.MYSQL_CONDITION_OP_GT,
                constant.MYSQL_CONDITION_OP_EGT,
                constant.MYSQL_CONDITION_OP_LT,
                constant.MYSQL_CONDITION_OP_ELT,
                constant.MYSQL_CONDITION_OP_NEQ,
                constant.MYSQL_CONDITION_OP_NULL_SAFE_EQUAL,
        ).__contains__(self._operate):
            return f"{self.get_field_expression()} {self._operate} " + MySQLKit.quote_offline(self._value)

        elif (
                constant.MYSQL_CONDITION_OP_IS,
                constant.MYSQL_CONDITION_OP_IS_NOT,
        ).__contains__(self._operate):
            if (
                    constant.MYSQL_CONDITION_CONST_TRUE,
                    constant.MYSQL_CONDITION_CONST_FALSE,
                    constant.MYSQL_CONDITION_CONST_NULL,
            ).__contains__(self._value):
                return f"{self.get_field_expression()} {self._operate} {self._value}"
            else:
                raise Exception("YOU MUST USE CONSTANT FOR IS COMPARISON!")

        elif (
                constant.MYSQL_CONDITION_OP_IN,
                constant.MYSQL_CONDITION_OP_NOT_IN,
        ).__contains__(self._operate):
            if type(self._value) is list or type(self._value) is tuple:
                if len(self._value) <= 0:
                    raise Exception("ARRAY IS EMPTY!")
                x = []
                for item in self._value:
                    x.append(MySQLKit.quote_offline(item))
                return f"{self.get_field_expression()} {self._operate} ({','.join(x)})"
            else:
                raise Exception("YOU MUST USE AN ARRAY FOR IN!")

        elif (
                constant.MYSQL_CONDITION_OP_LIKE,
                constant.MYSQL_CONDITION_OP_NOT_LIKE,
        ).__contains__(self._operate):
            return f"{self.get_field_expression()} {self._operate} {self._value}"

        elif (
                constant.MYSQL_CONDITION_OP_BETWEEN,
                constant.MYSQL_CONDITION_OP_NOT_BETWEEN,
        ).__contains__(self._operate):
            if type(self._value) is list or type(self._value) is tuple:
                if len(self._value) < 2:
                    raise Exception("NOT ENOUGH VALUES TO TEST IN BETWEEN!")
                left = MySQLKit.quote_offline(self._value[0])
                right = MySQLKit.quote_offline(self._value[1])
                return f"{self.get_field_expression()} {self._operate} {left} AND {right}"
            else:
                raise Exception("YOU MUST USE AN ARRAY FOR BETWEEN!")

        elif (
                constant.MYSQL_CONDITION_OP_EXISTS,
                constant.MYSQL_CONDITION_OP_NOT_EXISTS,
        ).__contains__(self._operate):
            # only value is used, as raw sql
            return f"{self._operate} ({self._value})"

        elif (
                constant.MYSQL_CONDITION_OP_PARENTHESES_AND,
                constant.MYSQL_CONDITION_OP_PARENTHESES_OR,
        ).__contains__(self._operate):
            if isinstance(self._value, (tuple, list)):
                if len(self._value) <= 0:
                    raise Exception("CONDITION ARRAY EMPTY!")
                parts = []
                for sub_condition in self._value:
                    if isinstance(sub_condition, MySQLCondition):
                        parts.append(sub_condition.organize_to_sql())
                    else:
                        raise Exception("NOT A MySQLCondition INSTANCE!")
                if len(parts) <= 0:
                    raise Exception("EMPTY CONDITION ARRAY!")

                return "(" + (" " + self._operate + " ").join(parts) + ")"
            else:
                raise Exception("YOU MUST USE AN ARRAY FOR CONDITIONS!")

        elif self._operate == constant.MYSQL_CONDITION_MACRO_IS_NULL_OR_EMPTY_STRING:
            return f"({self.get_field_expression()} IS NULL OR {self.get_field_expression()} = '')"

        elif self._operate == constant.MYSQL_CONDITION_MACRO_IS_NOT_NULL_NOR_EMPTY_STRING:
            return f"({self.get_field_expression()} IS NOT NULL AND {self.get_field_expression()} <> '')"

        elif self._operate == constant.MYSQL_CONDITION_MACRO_RAW_EXPRESSION:
            return self._value

        else:
            raise NotImplementedError("CANNOT OPERATE UNKNOWN TYPE!")

    @staticmethod
    def build_sql_component(conditions: Iterable["MySQLCondition"]):
        parts = []
        # Fixed in 0.1.9: If conditions have type error... RAISE!
        # The fucking Python design and PyCharm cannot deal with such check.
        for condition in conditions:
            if not isinstance(condition, MySQLCondition):
                raise TypeError("You must give an instance of MySQLCondition here!")
            parts.append(condition.organize_to_sql())
        # if not follow the rule, just ignore
        if len(parts) <= 0:
            return "1=1"
        else:
            return " AND ".join(parts)




LOGGER_CONFIG = {
    'log_dir': '/Users/sinri/code/nehushtan/log',
}
IMAP_CONFIG = {

}

MYSQL_CONFIG = {

}

SMTP_CONFIG = {

}

ALIYUN_SLS_CONFIG = {
    'project': 'zechariah',
    'logstore': 'zechariah-test',
    # 'endpoint': 'cn-hangzhou-intranet.log.aliyuncs.com',
    'endpoint': 'cn-hangzhou.log.aliyuncs.com',
    'source': '[IP]',
    'accessKeyId': 'XXX',
    'accessKeySecret': 'YYY',
}



from nehushtan.httpd.exceptions.NehushtanHTTPError import NehushtanHTTPError


class NehushtanRequestProcessTargetError(NehushtanHTTPError):
    """
    Since 0.4.0
    When the process target of the matched Route (filters or controllers) does not work
    """
    pass



from datetime import datetime

from nehushtan.mail.SMTPAgent import SMTPAgent
from tests.config import SMTP_CONFIG

now = datetime.now()

sm = SMTPAgent(
    sender=SMTP_CONFIG['sender'],
    password=SMTP_CONFIG['password'],
    smtp_server=SMTP_CONFIG['smtp_server'],
    smtp_port=SMTP_CONFIG['smtp_port']
) \
    .reset_receivers(['no-reply@anonymous.com']) \
    .set_subject('Test SMTP Agent') \
    .set_content(f'<h1>SMTP Agent</h1><p>It is a test mail</p><p>On {now}</p>') \
    .set_attachments(['test-for-smtp.py']) \
    .send()



import json
from typing import List, Optional

from nehushtan.score.ParsedLine import ParsedLine


class ParsedLineAsTitle(ParsedLine):

    def __init__(self, components: List[str]):
        super().__init__()

        self.__left_component: Optional[str] = None
        self.__middle_component: Optional[str] = None
        self.__right_component: Optional[str] = None

        if components is None or len(components) == 0:
            raise SyntaxError("title line components empty")
        if len(components) == 1:
            self.__middle_component = components[0]
        elif len(components) == 2:
            self.__left_component = components[0]
            self.__right_component = components[1]
        else:
            self.__left_component = components[0]
            self.__middle_component = components[1]
            self.__right_component = components[2]

    def get_left_component(self):
        if self.__left_component == "":
            return None
        return self.__left_component

    def get_middle_component(self):
        if self.__middle_component is None:
            return ""
        return self.__middle_component

    def get_right_component(self):
        if self.__right_component == "":
            return None
        return self.__right_component

    def __str__(self):
        x = (self.__left_component, self.__middle_component, self.__right_component,)
        return "ParsedLineAsTitle: " + json.dumps(x)



class CSVReaderRowHandleResult:
    def __init__(self):
        self.__should_continue = True

    def set_should_continue(self, should_continue: bool):
        self.__should_continue = should_continue
        return self

    def get_should_continue(self):
        return self.__should_continue



import base64
import hashlib
import random
import time
from typing import Tuple


class TOTPKit:
    def __init__(
            self,
            shared_secret: bytes,
            otp_length: int = 6,
            otp_seconds: int = 30
    ):
        self.__hash_algo = "sha1"
        self.__B = 64
        # every 30 seconds
        self.__F = otp_seconds
        self.__shared_secret = shared_secret  # b'BASE32SECRET2345AB=='
        # OTP Length
        self.__Digits = otp_length
        self.__key = self.__generate_key()

    @staticmethod
    def generate_shared_secret_with_base32() -> bytes:
        """
        Generate 20 bytes with Base32 supported.

        The most widely used Base32 alphabet is defined in RFC 4648.
        It uses an alphabet of A–Z, followed by 2–7.
        The digits 0, 1 and 8 are skipped due to their similarity with the letters O, I and B
        (thus "2" has a decimal value of 26).
        """

        shared_secret = []
        d = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
        # d = 'abcdefghijklmnopqrstuvwxyz234567'
        for i in range(20):
            b = d[random.Random().randint(0, len(d) - 1)]
            shared_secret.append(b.encode())
        return b''.join(shared_secret)

    @staticmethod
    def generate_key_with_base_32(shared_secret: bytes) -> bytes:
        """
        Google Authenticator Compatility (BASE-32)
        """
        return base64.b32decode(shared_secret + b'=' * (8 - len(shared_secret) % 8))

    def __generate_key(self) -> bytes:
        return TOTPKit.generate_key_with_base_32(self.__shared_secret)

    def __my_hmac(self, key: bytes, message: bytes) -> bytes:
        trans_5c = bytes((x ^ 0x5C) for x in range(256))
        trans_36 = bytes((x ^ 0x36) for x in range(256))
        k_zpad = key.ljust(self.__B, b'\0')
        k_ipad = k_zpad.translate(trans_36)
        k_opad = k_zpad.translate(trans_5c)
        hash1 = hashlib.new(self.__hash_algo, k_ipad + message).digest()
        hmac_hash = hashlib.new(self.__hash_algo, k_opad + hash1).digest()
        return hmac_hash

    def __dynamic_truncate(self, b_hash) -> int:
        hash_len = len(b_hash)
        int_hash = int.from_bytes(b_hash, byteorder='big')
        offset = int_hash & 0xF
        # Geterate a mask to get bytes from left to right of the hash
        n_shift = 8 * (hash_len - offset) - 32
        mask = 0xFFFFFFFF << n_shift
        hex_mask = "0x" + ("{:0" + str(2 * hash_len) + "x}").format(mask)
        p = (int_hash & mask) >> n_shift  # Get rid of left zeros
        lsb_31 = p & 0x7FFFFFFF  # Return only the lower 31 bits
        return lsb_31

    def __generate_totp_tuple(self, apply_time: int) -> Tuple[int, int, int, str]:
        """
        With an apply time, generate TOTP for it.

        Return a tuple with:
            apply_time
            available_time
            expire_time
            TOTP
        """
        # %30 seconds
        formated_time = int(apply_time / self.__F)
        t = formated_time.to_bytes(8, byteorder='big')
        # Same algorithm as HOTP
        hmac_hash = self.__my_hmac(self.__key, t)
        trc_hash = self.__dynamic_truncate(hmac_hash)  # Get truncated hash (int)
        # Adjust totp length
        totp = ("{:0" + str(self.__Digits) + "}").format(trc_hash % (10 ** self.__Digits))
        return apply_time, formated_time * self.__F, formated_time * self.__F + self.__F, totp

    def generate_current_totp_tuple(self) -> Tuple[int, int, int, str]:
        """
        Generate TOTP for current time.

        Return a tuple with:
            apply_time
            available_time
            expire_time
            TOTP
        """
        t0 = int(time.time())
        totp_now = self.__generate_totp_tuple(t0)
        return totp_now



import os
import time
from multiprocessing import Pool


def work(name):
    print(time.time(), name)
    return f'{name} worked on {time.time()} pid={os.getpid()}'


if __name__ == '__main__':
    with Pool(processes=3) as pool:
        res = pool.apply_async(func=work, args=['A'])
        x = res.get(timeout=1)
        print(x)
    print(f'Fin on pid={os.getpid()}')



from abc import abstractmethod

import PySimpleGUI as PySimpleGUI

from nehushtan.gui.simple.constants.SimpleGuiTheme import SimpleGuiTheme


class SimpleWindow:
    def __init__(
            self,
            title: str,
            theme: SimpleGuiTheme = None
    ):
        if theme is not None:
            # Add a touch of color
            PySimpleGUI.theme(theme.get_theme_name())

        # Create the Window
        self.__window = PySimpleGUI.Window(title)

    def get_window(self):
        return self.__window

    @abstractmethod
    def get_current_layout(self) -> list[list]:
        pass

    def set_window_layout(self, layout: list[list]):
        # All the stuff inside your window.
        self.__window.layout(layout)
        return self

    @abstractmethod
    def should_continue_loop_after_event(self, event, values) -> bool:
        """
        if return false, exit event loop
        """
        pass

    def start_loop(self):
        # initialize layout
        self.set_window_layout(self.get_current_layout())
        # Event Loop to process "events" and get the "values" of the inputs
        while True:
            event, values = self.__window.read()
            print(event, values)
            # if event == PySimpleGUI.WIN_CLOSED \
            #         or event == 'Cancel' or event == 'Ok':
            #     # if user closes window or clicks cancel
            #     break
            # print('You entered ', values[0])
            if not self.should_continue_loop_after_event(event, values):
                break

        self.__window.close()



import time

from nehushtan.helper.timer.TimeCostResult import TimeCostResult


class TimeCostHelper:
    def __init__(self, result: TimeCostResult):
        self.result = result

    def __enter__(self):
        self.result.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.result.end = time.time()
        self.result.period = self.result.end - self.result.start


if __name__ == '__main__':
    """
    This is an example
    """

    tcr = TimeCostResult()
    with TimeCostHelper(tcr):
        time.sleep(2)
    print(tcr)



import time
from random import random

from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLoggerAdapterWithFileWriter
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask


class Test2NehushtanQueueTask(NehushtanQueueTask):
    def __init__(self, task_id: int):
        super().__init__()

        self._task_id = task_id
        self.logger = NehushtanLogger(
            topic=f'task-{self._task_id}',
            adapter=NehushtanLoggerAdapterWithFileWriter(log_dir='/Users/leqee/code/nehushtan/log/queue-plan2/task',
                                                         categorize=False,
                                                         date_rotate=False
                                                         ),
        )

    def get_task_reference(self):
        return self._task_id

    def get_task_type(self):
        return 'TEST_TASK'

    def execute(self):
        time_cost = int(10 * random())
        self.logger.info(f'Task Execute Plan to spend {time_cost} seconds', {'task_id': self.get_task_reference()})
        time.sleep(time_cost)
        self.execute_result = random()
        if self.execute_result > 0.9:
            raise RuntimeError('Boom!')
        elif self.execute_result > 0.8:
            self.done = False
        else:
            self.done = True

        self.execute_feedback = f'Result is {self.execute_result}, and Done is {self.done}'
        self.logger.info(f'Task Executed', {'task_id': self.get_task_reference()})



import unittest

from nehushtan.mysql.MySQLAnyTable import MySQLAnyTable
from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from nehushtan.mysql.MySQLUnionSelection import MySQLUnionSelection
from tests.config import MYSQL_CONFIG


class TestForMySQLSelection(unittest.TestCase):
    def setUp(self) -> None:
        self.table = MySQLAnyTable(
            MySQLKit(MySQLKitConfig(MYSQL_CONFIG)),
            'c'
        )

        self.table.truncate()
        for i in range(100):
            self.table.insert_one_row({'value': i * i, 'name': f'item [{i}]'})

    def test_1(self):
        result = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 100, 200)) \
            .query_for_result_as_tuple_of_dict()

        the_tuple = result.get_fetched_rows_as_tuple()
        print(len(the_tuple))
        for i in the_tuple:
            print(i)

    def test_2(self):
        result = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 400, 800)) \
            .query_for_result_as_tuple_of_dict()

        the_tuple = result.get_fetched_rows_as_tuple()
        print(len(the_tuple))
        for i in the_tuple:
            print(i)

    def test_3(self):
        s1 = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 100, 200))
        s2 = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 400, 800))

        result = MySQLUnionSelection() \
            .add_sub_selection(s1) \
            .add_sub_selection(s2) \
            .query_for_result_as_tuple_of_dict()

        print(result.get_sql())
        print(result.get_error())

        the_tuple = result.get_fetched_rows_as_tuple()
        print(len(the_tuple))
        for i in the_tuple:
            print(i)

    def test_4(self):
        s1 = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 100, 500))
        s2 = self.table.select_in_table() \
            .add_condition(MySQLCondition.make_between('value', 200, 800))

        result = MySQLUnionSelection() \
            .add_sub_selection(s1) \
            .add_sub_selection(s2, True) \
            .query_for_result_as_tuple_of_dict()

        print(result.get_sql())
        print(result.get_error())

        the_tuple = result.get_fetched_rows_as_tuple()
        print(len(the_tuple))
        for i in the_tuple:
            print(i)



import signal
import time
from multiprocessing import Process
from typing import List

from nehushtan.helper.SignalHandler import SignalHandler
from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLogLevel
from nehushtan.multiprocessing.NehushtanMPJob import NehushtanMPJob
from nehushtan.multiprocessing.NehushtanMPTerminatedSituation import NehushtanMPTerminatedSituation


class NehushtanMPQueue(SignalHandler):
    """
    Since 0.2.13

    Notice:
    The program entrance should be put into `if __name__ == '__main__':`;
    and need `set_start_method('spawn')` in the beginning.
    """

    def __init__(self):
        self._logger = None

        self.__max_workers = 1
        self.__waiting_queue = []
        self.__running_dict = {}
        self.__running_worker_dict = {}

        self.__received_terminate_signal = False

    def set_logger(self, logger: NehushtanLogger):
        self._logger = logger
        return self

    def get_logger(self) -> NehushtanLogger:
        if self._logger is None:
            self._logger = NehushtanLogger(log_level=NehushtanLogLevel.NOTICE)
        return self._logger

    def get_max_workers(self):
        return self.__max_workers

    def set_max_workers(self, max_workers: int):
        self.__max_workers = max_workers
        return self

    def get_target_signal_list(self) -> List[int]:
        return [signal.SIGTERM]

    def handle_signal(self, signal_number, frame):
        """
        Handle SIGNAL 15 (TERMINATE) as Callback
        Let all processes in multiprocessing receive SIGTERM signal
        """

        self.__received_terminate_signal = True

        self.get_logger().warning('NehushtanMPQueue::handle_signal received signal', signal_number)
        if signal_number == signal.SIGTERM:
            self.__waiting_queue = []
            self.get_logger().critical('NehushtanMPQueue::handle_signal cleaned the waiting queue')
            for pid, p in self.__running_dict.items():
                if p.is_alive():
                    self.get_logger().critical(
                        'NehushtanMPQueue::handle_signal terminate alive sub process',
                        {'pid': p.pid, 'is_alive': p.is_alive(), }
                    )
                    p.terminate()
                else:
                    self.get_logger().critical(
                        'NehushtanMPQueue::handle_signal terminate sub process exited already',
                        {'pid': p.pid, 'is_alive': p.is_alive(), 'exit_code': p.exitcode, }
                    )

    def enqueue_job(self, job: NehushtanMPJob):
        self.__waiting_queue.insert(0, job)
        self.get_logger().info('NehushtanMPQueue::enqueue_job', job.get_name())
        return self

    def __create_worker_for_job(self, job: NehushtanMPJob):
        p = Process(target=job.handle)
        p.start()
        job.set_pid(p.pid)
        self.__running_dict[p.pid] = p
        self.__running_worker_dict[p.pid] = job
        self.get_logger().info(
            'NehushtanMPQueue::__create_worker_for_job',
            {'job': job.get_name(), 'pid': job.get_pid()}
        )
        return p.pid

    def execute_all_jobs(self, rest_seconds=5, max_workers: int = None):
        """
        The NehushtanMPQueue starts after all jobs were enqueued.
        This method would clean
        """

        # register signal handler for SIGTERM
        self.apply()

        if max_workers is not None:
            self.set_max_workers(max_workers)

        self.get_logger().info('NehushtanMPQueue::execute_all_jobs', {'pool_size': self.get_max_workers()})
        while True:
            current_process_count = self.scan_workers()
            waiting_jobs = len(self.__waiting_queue)
            self.get_logger().info(
                'NehushtanMPQueue::execute_all_jobs report',
                {'running_jobs': current_process_count, 'waiting_jobs': waiting_jobs, 'workers': self.get_max_workers()}
            )

            if current_process_count <= 0:
                # multiprocessing empty, check waiting queue
                if waiting_jobs <= 0:
                    # if no waiting task, ends
                    break
                job = self.__waiting_queue.pop()
                self.__create_worker_for_job(job)

            elif current_process_count < self.get_max_workers():
                # multiprocessing has free space, check waiting queue
                if waiting_jobs <= 0:
                    # if no waiting task, sleep
                    time.sleep(rest_seconds)
                else:
                    job = self.__waiting_queue.pop()
                    self.__create_worker_for_job(job)

            else:
                # multiprocessing is full, sleep
                time.sleep(rest_seconds)

        self.get_logger().info('NehushtanMPQueue::execute_all_jobs all sub-processes ended')

        # unregister signal handler for SIGTERM, replaced by the default
        self.apply_default()

        if self.__received_terminate_signal:
            raise NehushtanMPTerminatedSituation(
                f'NehushtanMPQueue received SIGTERM during running execute_all_jobs.'
            )

    def when_worker_observed_exited(self, pid: int, exit_code: int):
        """
        When worker (sub process) is found as exited in `scan_workers`,
        this method would be called, normally in main process.
        """
        self.get_logger().info(
            'NehushtanMPQueue::when_worker_observed_exited',
            {'pid': pid, 'exit_code': exit_code}
        )
        worker = self.__running_worker_dict.get(pid)
        if worker:
            worker.when_exited(exit_code)

    def scan_workers(self) -> int:
        """
        Scan all workers in `__running_dict`,
        Clear the exited ones and report,
        Then return the count of the remaining.
        """
        if len(self.__running_dict.items()) <= 0:
            return 0

        finished_pid = []
        for pid, p in self.__running_dict.items():
            if not p.is_alive():
                self.get_logger().info(
                    'NehushtanMPQueue::refresh_running_sub_process_count a sub process is not alive now',
                    {'pid': p.pid, 'exit_code': p.exitcode}
                )
                finished_pid.append(pid)
                self.when_worker_observed_exited(p.pid, p.exitcode)

        for pid in finished_pid:
            del self.__running_dict[pid]

        self.get_logger().info(
            'NehushtanMPQueue::refresh_running_sub_process_count now still alive process',
            {'keys': self.__running_dict.keys()}
        )

        return len(self.__running_dict.items())



import quopri
import re
from base64 import b64decode, b64encode


class EncodedWordsKit:
    """
    Since 0.4.6
    """

    @staticmethod
    def decode_string_following_rfc2047(raw_string: str):
        """
        See https://datatracker.ietf.org/doc/html/rfc2047
        See https://tools.ietf.org/html/rfc2045
        Simply See https://dmorgan.info/posts/encoded-word-syntax/

        Encoded-word Syntax
        encoded-word = "=?" charset "?" encoding "?" encoded-text "?="

        `charset` is one of the character sets registered with IANA for use
            with the MIME text/plain content-type
        `encoding` is either the character Q or B,
            where B represents base64 encoding
            and Q represents an encoding similar to the “quoted-printable” content transfer encoding defined in RFC 2045
        `encoded-text` is the text encoded according to the defined `encoding`

        Sample:
        "=?UTF-8?B?5oqW5bqX?=" or =?UTF-8?B?5oqW5bqX?=
        Pared to 抖店
        """
        matched = re.match(r'"?=\?([A-Za-z0-9_-]+)\?([QB])\?(.+)\?="?', raw_string)
        if matched:
            charset = matched[1]
            encoding = matched[2]  # Q or B
            raw = matched[3]
            if encoding == 'B':
                return b64decode(raw).decode(charset)
            if encoding == 'Q':
                return quopri.decodestring(raw).decode(charset)
        return raw_string

    @staticmethod
    def encode_string_following_rfc2047(string: str, charset: str, encoding: str):
        """
        `charset` might be `UTF-8`
        `encoding` might be `Q` or `B`
        """
        byte_array = string.encode(charset)
        if encoding == 'B':
            x = b64encode(byte_array).decode()
        elif encoding == 'Q':
            x = quopri.encodestring(byte_array).decode()
        else:
            return string
        return f'=?{charset}?{encoding}?{x}?='



import json


class MQCommand:
    COMMAND_ENQUEUE = 'ENQUEUE'
    COMMAND_DEQUEUE = 'DEQUEUE'

    def __init__(self, command: str, queue_name: str, value: str = None):
        self.command = command
        self.queue_name = queue_name
        self.value = value

    @staticmethod
    def parse(data: bytes):
        decoded = data.decode()
        x = json.loads(decoded)
        if type(x) is not dict:
            return None
        command = x.get('command')
        queue_name = x.get('queue_name')
        value = x.get('value')

        if command != MQCommand.COMMAND_DEQUEUE and command != MQCommand.COMMAND_ENQUEUE:
            return None

        if type(queue_name) is not str or len(queue_name) <= 0:
            return None

        if command == MQCommand.COMMAND_ENQUEUE and value is None:
            return None

        return MQCommand(command, queue_name, value)

    def package_as_bytes(self) -> bytes:
        x = json.dumps({'command': self.command, 'queue_name': self.queue_name, 'value': self.value})
        return x.encode()



from nehushtan.httpd.NehushtanHTTPRequestFilter import NehushtanHTTPRequestFilter


class TestFilter(NehushtanHTTPRequestFilter):
    def should_accept_request(self) -> bool:
        """
        Sample
        ------
        URL: http://localhost:4466/api/TestApi/hi/there?limit=0

        print(self.get_http_handler().parsed_path) # /api/TestApi/hi/there
        print(self.get_http_handler().parsed_query_dict) # {'limit': '0'}
        print(self.get_http_handler().parsed_body_data) # None -> Not Parsed
        """
        if self.get_http_handler().parsed_query_dict.get('limit') == '0':
            return False
        return True



from email.message import EmailMessage, Message
from typing import Union, Optional, List


class NehushtanMessagePart:
    """
    Since 0.4.8
    """

    def __init__(self, msg: Union[EmailMessage, Message], parent: Optional['NehushtanMessagePart']):
        self.__parent = parent
        self.__message = msg
        self.__sub_message_list: List['NehushtanMessagePart'] = []

        self.__organize_parts()

    def get_parent(self):
        return self.__parent

    def get_message(self):
        return self.__message

    def get_sub_message_list(self):
        return self.__sub_message_list

    def is_leaf(self):
        return len(self.__sub_message_list) == 0

    # def append_sub_message(self, sub_message):
    #     self.__sub_message_list.append(sub_message)
    #     return self

    def __organize_parts(self):
        if self.__message.is_multipart():
            for part in self.__message.iter_parts():
                self.__sub_message_list.append(NehushtanMessagePart(part, self))

    def get_content_charset(self):
        return self.__message.get_content_charset()

    def get_content_type(self):
        return self.__message.get_content_type()

    def get_content_disposition(self):
        return self.__message.get_content_disposition()

    def get_body_content(self, i=None):
        x: bytes = self.__message.get_payload(i=i, decode=True)
        return x.decode(encoding=self.__message.get_content_charset('ascii'))

    def get_attachment_filename_of_this_part(self) -> Optional[str]:
        if self.__message.is_attachment():
            return self.__message.get_filename()

    def get_attachment_content_of_this_part(self, i=None):
        if self.__message.is_attachment():
            attachment_content = self.__message.get_payload(i=i, decode=True)
            return attachment_content

    def save_attachment_of_this_part(self, target_file_path: str, i=None):
        with open(target_file_path, 'wb') as fp:
            fp.write(self.get_attachment_content_of_this_part(i=i))



import sys

from PIL import Image, ImageDraw

if __name__ == '__main__':
    image = Image.new("RGB", (100, 200))
    draw = ImageDraw.Draw(image)
    draw.line((0, 0) + image.size, fill=128)
    draw.line((0, image.size[1], image.size[0], 0), fill=128)

    # write to stdout
    stdout_buffer = sys.stdout.buffer
    image.save(stdout_buffer, "PNG")

    image.show("aaa")

    image.close()



from typing import Optional, Union, List

from nehushtan.mysql.MySQLJoinOnCondition import MySQLJoinOnCondition
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLSelectionMixin import MySQLSelectionTarget, MySQLSelectionMixin
from nehushtan.mysql.MySQLTableExistence import MySQLTableExistence
from nehushtan.mysql.MySQLTableSelection import MySQLTableSelection


class MySQLJoinedTables(MySQLSelectionTarget):
    def __init__(self):
        super().__init__()
        self.__join_target_list = []
        self.__db: Optional[MySQLKit] = None

    def add_join_target(self,
                        join_type: str,
                        target: Union[MySQLSelectionMixin, MySQLTableExistence],
                        alias: Optional[str] = None,
                        on_conditions: Optional[List[MySQLJoinOnCondition]] = None,
                        index_hint: Optional[str] = None,
                        ):
        db = target.get_mysql_kit()
        if db is not None:
            self.__db = db

        x = join_type + " "
        if isinstance(target, MySQLTableExistence):
            x += target.get_table_expression()
        elif isinstance(target, MySQLSelectionMixin):
            x += target.generate_sql()
        else:
            raise TypeError(
                'target for nehushtan.mysql.MySQLJoinSelection.MySQLJoinSelection.add_join_target should not be this')

        if alias is not None:
            x += ' as `' + alias + '`'

        if index_hint is not None:
            x += ' ' + index_hint

        if on_conditions is not None and len(on_conditions) > 0:
            a = []
            for condition in on_conditions:
                a.append(condition.organize_to_sql())
            x += ' on ' + ' and '.join(a)

        self.__join_target_list.append(x)
        return self

    def based_on(self,
                 target: Union[MySQLSelectionMixin, MySQLTableExistence],
                 alias: Optional[str] = None,
                 index_hint: Optional[str] = None,
                 ):
        return self.add_join_target(join_type='',
                                    target=target, alias=alias,
                                    on_conditions=None,
                                    index_hint=index_hint)

    def left_join(self,
                  target: Union[MySQLSelectionMixin, MySQLTableExistence],
                  alias: Optional[str] = None,
                  index_hint: Optional[str] = None,
                  on_conditions: Optional[List[MySQLJoinOnCondition]] = None,
                  ):
        return self.add_join_target(join_type='LEFT JOIN',
                                    target=target, alias=alias,
                                    on_conditions=on_conditions,
                                    index_hint=index_hint)

    def inner_join(self,
                   target: Union[MySQLSelectionMixin, MySQLTableExistence],
                   alias: Optional[str] = None,
                   index_hint: Optional[str] = None,
                   on_conditions: Optional[List[MySQLJoinOnCondition]] = None,
                   ):
        return self.add_join_target(join_type='INNER JOIN',
                                    target=target, alias=alias,
                                    on_conditions=on_conditions,
                                    index_hint=index_hint)

    def get_mysql_kit(self) -> MySQLKit:
        return self.__db

    def as_selection_target(self) -> str:
        return ' '.join(self.__join_target_list)

    def select_in_joined_tables(self):
        return MySQLTableSelection(target=self, )


import asyncio
from abc import abstractmethod
from typing import List

from websockets.exceptions import ConnectionClosedError

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.ws.NehushtanWebsocketConnectionEntity import NehushtanWebsocketConnectionEntity


class NehushtanWebsocketAgent:
    def __init__(
            self,
            broadcast_sleep_seconds: int,
            request_handle_logger: NehushtanLogger,
            broadcast_logger: NehushtanLogger
    ):
        self.__request_handle_logger = request_handle_logger
        self.__broadcast_logger = broadcast_logger
        self.__local_key = ''
        self.broadcast_sleep_seconds = broadcast_sleep_seconds

    def get_local_key(self):
        return self.__local_key

    def make_key_of_websocket(self, websocket) -> str:
        local_key = f'{websocket.local_address[0]}-{websocket.local_address[1]}'
        remote_key = f'{websocket.remote_address[0]}-{websocket.remote_address[1]}'

        if self.__local_key == '':
            self.__local_key = local_key
        elif self.__local_key != local_key:
            raise RuntimeError(f"NehushtanWebsocketAgent LOCAL KEY CHANGED {self.__local_key} -> {local_key}")

        return f'{local_key}~{remote_key}'

    @abstractmethod
    def register_raw_websocket_connection(self, websocket):
        """
        Save the argument `websocket`,
        and bind it to its key, which should be computed by `self.make_key_of_websocket(websocket)`
        """
        pass

    @abstractmethod
    def unregister_raw_websocket_connection(self, websocket_key: str):
        """
        Remove the connection from the registration according to its key `websocket_key`
        """
        pass

    @abstractmethod
    def build_response_content(self, websocket_key: str, received_content: str) -> str:
        """
        Handle the message and build the response content to send back
        """
        pass

    async def ws_request_handler(self, websocket, path):
        while True:
            connection_key = self.make_key_of_websocket(websocket)
            self.__request_handle_logger.debug(f"Handling connection [{connection_key}] for path [{path}]")
            self.register_raw_websocket_connection(websocket)

            try:
                received_content = await websocket.recv()
                self.__request_handle_logger.info(f"Received From [{connection_key}]", {'content': received_content})

                response_content = self.build_response_content(connection_key, received_content)
                if len(response_content) > 0:
                    await websocket.send(response_content)
                    self.__request_handle_logger.info(f"Sent to [{connection_key}]", {'content': response_content})
            except ConnectionClosedError as theConnectionClosedError:
                self.__request_handle_logger.info(
                    f"Connection [{connection_key}] lost, to remove. Error: {theConnectionClosedError}"
                )
                self.unregister_raw_websocket_connection(connection_key)
                break

    @abstractmethod
    def get_connections_related_to_this_agent(self) -> List[NehushtanWebsocketConnectionEntity]:
        pass

    @abstractmethod
    def check_content_to_send_to_target_client(self, connection: NehushtanWebsocketConnectionEntity) -> str:
        pass

    async def ws_broadcast_handler(self):
        while True:
            connections = self.get_connections_related_to_this_agent()

            for connection in connections:
                self.__broadcast_logger.debug(f"Broadcast to [{connection.get_key()}]")
                try:
                    await connection.get_websocket().ping()

                    content_to_send = self.check_content_to_send_to_target_client(connection)
                    if content_to_send:
                        await connection.get_websocket().send(content_to_send)
                except ConnectionClosedError as theConnectionClosedError:
                    self.__broadcast_logger.info(
                        f"Connection [{connection.get_key()}] lost, to remove. Error: {theConnectionClosedError}"
                    )
                    self.unregister_raw_websocket_connection(connection.get_key())

            await asyncio.sleep(self.broadcast_sleep_seconds)



from typing import Tuple, Union


class Note:
    """
    0: C
    1: C# = Db
    2: D
    3: D# = Eb
    4: E
    5: F
    6: F# = Gb
    7: G
    8: G# = Ab
    9: A
    10: A# = Bb
    11: B
    """

    def __init__(self, name: str, group: int = 4, move_half: int = 0, beats: Union[int, float] = 1):
        """
        C = C4
        D 5 1 = D5# = E5b
        B 4 -1 = B4b = A4#
        """
        self.group = group
        self.name = name
        self.flat = move_half < 0
        self.sharp = move_half > 0
        self.__beats = beats

    def __as_indexed(self):
        a = self.group
        if self.name == 'C':
            b = 0
        elif self.name == 'D':
            b = 2
        elif self.name == 'E':
            b = 4
        elif self.name == 'F':
            b = 5
        elif self.name == 'G':
            b = 7
        elif self.name == 'A':
            b = 9
        elif self.name == 'B':
            b = 11
        else:
            raise RuntimeError("unknown name")

        if self.flat:
            if b == 0:
                b = 11
                a = a - 1
            else:
                b = b - 1
        elif self.sharp:
            if b == 11:
                b = 0
                a = a + 1
            else:
                b = b + 1
        return a, b,

    @staticmethod
    def __make_note_from_indexed(indexed: Tuple[int, int]):
        group = indexed[0]
        b = indexed[1]
        move_half = 0

        if b == 0:
            name = 'C'
        elif b == 1:
            name = 'D'
            move_half = -1
        elif b == 2:
            name = 'D'
        elif b == 3:
            name = 'E'
            move_half = -1
        elif b == 4:
            name = 'E'
        elif b == 5:
            name = 'F'
        elif b == 6:
            name = 'G'
            move_half = -1
        elif b == 7:
            name = 'G'
        elif b == 8:
            name = 'A'
            move_half = -1
        elif b == 9:
            name = 'A'
        elif b == 10:
            name = 'B'
            move_half = -1
        elif b == 11:
            name = 'B'
        else:
            raise RuntimeError

        return Note(name, group, move_half)

    def get_beats(self):
        return self.__beats

    def get_note_higher(self, distance: int):
        if distance < 0:
            return self.get_note_lower(abs(distance))
        t = self.__as_indexed()
        a = t[0]
        b = t[1]
        for i in range(distance):
            b = b + 1
            if b == 12:
                b = 0
                a = a + 1
        return Note.__make_note_from_indexed((a, b))

    def get_note_lower(self, distance: int):
        if distance < 0:
            return self.get_note_higher(abs(distance))
        t = self.__as_indexed()
        a = t[0]
        b = t[1]
        for i in range(distance):
            if b == 0:
                b = 11
                a = a - 1
            else:
                b = b - 1
        return Note.__make_note_from_indexed((a, b))

    def get_distance_to_another_note(self, another_note):
        """
        X = self - another
        """
        a = self.__as_indexed()
        b = another_note.__as_indexed()

        return (a[0] - b[0]) * 12 + (a[1] - b[1])

    def __get_distance_to_a4(self):
        a = self.__as_indexed()
        return (a[0] - 4) * 12 + (a[1] - 9)

    def get_sound_in_hz(self) -> float:
        """
        f(A4) = 440 Hz
        For any other note X, n=X-A4 (1 for half), f(X)=2^(n/12)*440 Hz
        For Example, if X is C5, n=3 (A4-#-B4-C5), f(C5)=2^(3/12)*440=523.2511306011972 Hz
        """
        distance_to_a4 = self.__get_distance_to_a4()
        return pow(2, distance_to_a4 / 12.0) * 440.0



from nehushtan.httpd.exceptions.NehushtanHTTPError import NehushtanHTTPError


class NehushtanNoRouteMatchedError(NehushtanHTTPError):
    """
    Since 0.4.0
    When the Router cannot find a route to match the incoming request
    """

    def __init__(self, http_error_message: str, http_code: int):
        self.http_error_message = http_error_message
        self.http_code = http_code
        super().__init__(http_error_message, http_code)



from nehushtan.helper.io.CSVReader import CSVReader
from nehushtan.helper.io.CSVReaderOptions import CSVReaderOptions
from nehushtan.helper.io.CSVReaderRowHandleResult import CSVReaderRowHandleResult

csv_file = "/Users/leqee/code/nehushtan/debug/attachments/1.csv"


# def row_transformer(row, index: int, field_names: Sequence[str]) -> any:
#     rest_key = None
#     rest_val = None
#     rest_list = []
#
#     d = dict()
#     for i in range(max(len(row), len(field_names))):
#         if i < len(row):
#             if i < len(field_names):
#                 d[field_names[i]] = row[i]
#             else:
#                 rest_list.append(row[i])
#         else:
#             d[field_names[i]] = rest_val
#
#     if len(rest_list) > 0:
#         d[rest_key] = rest_list
#
#     return d


def row_handler_1(row, index: int, result: CSVReaderRowHandleResult):
    if index == 2 or index == 4:
        result.set_should_continue(False)
    print(index, row, result.get_should_continue())


if __name__ == '__main__':
    csv_reader = CSVReader(csv_file, CSVReaderOptions())

    csv_reader.set_row_transformer(CSVReader.basic_row_to_dict_transformer)
    csv_reader.set_row_handler(row_handler_1)
    csv_reader.take_first_row_as_field_names()

    print("from", csv_reader.next_line_index())
    csv_reader.handle_next_line()
    print("to", csv_reader.next_line_index())

    print("from", csv_reader.next_line_index())
    csv_reader.handle_next_lines()
    print("to", csv_reader.next_line_index())

    print("from", csv_reader.next_line_index())
    csv_reader.handle_next_line()
    print("to", csv_reader.next_line_index())

    print("from", csv_reader.next_line_index())
    csv_reader.handle_next_lines()
    print("to", csv_reader.next_line_index())

    print("from", csv_reader.next_line_index())
    csv_reader.handle_next_line()
    print("to", csv_reader.next_line_index())



class PgSQLKitConfig:
    """
    Since 0.4.22
    """

    def __init__(self, config_dict: dict):
        # host – database host address (defaults to UNIX socket if not provided)
        # port – connection port number (defaults to 5432 if not provided)
        # dbname - the database name (database is a deprecated alias)
        # user – user name used to authenticate
        # password – password used to authenticate
        # autocommit - BOOL
        self.__config_dict = config_dict

    def dsn(self):
        dsn = []
        for k in ('host', 'port', 'dbname', 'user', 'password'):
            v = self.__config_dict.get(k)
            if v:
                dsn.append(f'{k}={v}')
        return ' '.join(dsn)

    def autocommit(self) -> bool:
        return self.__config_dict.get('autocommit') is True

    def __str__(self):
        return self.dsn()



import re

from nehushtan.score.ParsedLineAsBlank import ParsedLineAsBlank
from nehushtan.score.ParsedLineAsLyric import ParsedLineAsLyric
from nehushtan.score.ParsedLineAsScore import ParsedLineAsScore
from nehushtan.score.ParsedLineAsTitle import ParsedLineAsTitle
from nehushtan.score.ScoreUnit import ScoreUnit


class NotationTextParser:
    def __init__(self, text: str = None, file: str = None):
        self.__lines = []
        self.__parsed_lines = []
        if text is not None:
            lines = text.splitlines(keepends=False)
            for line in lines:
                self.__lines.append(line)
        elif file is not None:
            with open(file, "r", encoding="utf-8") as f:
                lines = f.readlines()
                self.__text = ""
                for line in lines:
                    self.__lines.append(line)
        else:
            pass

    def parse(self):
        for line in self.__lines:
            if line is None or line == "" or len(line.strip()) == 0:
                self.__parsed_lines.append(ParsedLineAsBlank())
            elif line.startswith("~"):
                self.__parse_title_line(line)
            elif line.startswith(">"):
                self.__parse_lyric_line(line)
            else:
                self.__parse_score_line(line)

    def __parse_title_line(self, line):
        compiled = re.compile("^~ (.+)\s*$")
        r = compiled.match(line)
        if r:
            x = r.group(1)
            components = re.compile("\s*\|\s*").split(x)
            self.__parsed_lines.append(ParsedLineAsTitle(components))

    def __parse_lyric_line(self, line):
        compiled = re.compile("^>((.+)([<=>]))? (.+)\s*$")
        r = compiled.match(line)
        if r:
            prefix = r.group(2)
            if prefix == "":
                prefix = None
            content_align_type = r.group(3)
            if content_align_type not in ("<", "="):
                content_align_type = ">"
            content = r.group(4)

            # if len(r.groups()) == 2:
            #     prefix = None
            #     content = r.group(1)
            # else:
            #     prefix = r.group(2)
            #     content = r.group(3)
            self.__parsed_lines.append(ParsedLineAsLyric(content, prefix, content_align_type))

    def __parse_score_line(self, line):
        # print(line)
        compiled = re.compile("\s+")
        items = compiled.split(line.strip())
        score_unit_list = []
        for item in items:
            #print(json.dumps(item))
            score_unit = ScoreUnit.parse_code_to_score(item)
            score_unit_list.append(score_unit)
        self.__parsed_lines.append(ParsedLineAsScore(score_unit_list))

    def debug(self):
        for p in self.__parsed_lines:
            print(p)

    def get_parsed_score_lines(self):
        return self.__parsed_lines



import socket
import time
from typing import Optional

from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLoggerAdapterWithFileWriter
from nehushtan.socket.SocketHandleThreadManager import SocketHandlerThreadManager
from nehushtan.socket.wormhole.NehushtanWormholeWorker import NehushtanWormholeWorker


class NehushtanWormhole:
    """
    Since 0.4.28
    """

    def __init__(self, port: int, dest_host: str, dest_port: int, log_dir: str = None):
        self.__listen_host = '0.0.0.0'
        self.__listen_port = port
        self.__dest_host = dest_host
        self.__dest_port = dest_port

        self.__wait_queue_size = 0
        self.__socket_instance: Optional[socket.socket] = None

        self.__thread_manager = SocketHandlerThreadManager()

        # logger: NehushtanFileLogger
        # self.__logger = logger
        self.__log_dir = log_dir
        if self.__log_dir is None:
            self.__logger = NehushtanLogger(topic="Wormhole", )
        else:
            self.__logger = NehushtanLogger(topic="Wormhole",
                                            adapter=NehushtanLoggerAdapterWithFileWriter(
                                                log_dir=self.__log_dir,
                                            ))

    def set_logger(self, logger: NehushtanLogger):
        self.__logger = logger
        return self

    def run(self):
        address_entries = socket.getaddrinfo(
            self.__listen_host,
            self.__listen_port,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM,
            0,
            socket.AI_PASSIVE
        )

        prepare_error = None

        for address_entry in address_entries:
            af, socket_type, proto, canon_name, sa = address_entry
            try:
                self.__socket_instance = socket.socket(af, socket_type, proto)
                self.__socket_instance.bind(sa)
                self.__socket_instance.listen(self.__wait_queue_size)
            except OSError as msg:
                prepare_error = msg
                if self.__socket_instance is not None:
                    self.__socket_instance.close()
                self.__socket_instance = None
                continue
            break

        if self.__socket_instance is None:
            raise prepare_error

        # let the process of accept could be interrupted (Since 0.4.17)
        self.__socket_instance.settimeout(1)

        self.__logger.notice(f"Wormhole Entrance is open, liston on {self.__listen_port}")

        self.__serving_thread()

    def __serving_thread(self):
        worker_thread_index = 0
        previous_alive_count = 0
        while True:
            alive_count = self.__thread_manager.check_alive_thread_count()
            if previous_alive_count != alive_count:
                self.__logger.info(f"NOW ALIVE CLIENT COUNT IS {alive_count}")
                previous_alive_count = alive_count

            try:
                connection, address = self.__socket_instance.accept()
                self.__logger.info(f"Accept New Client: {address}")
            except socket.timeout:
                continue

            worker_thread_index += 1

            socket_name = f"{time.time()}-{worker_thread_index}"
            worker = NehushtanWormholeWorker(socket_name, connection, address, self.__dest_host, self.__dest_port,
                                             self.__log_dir, self.__thread_manager)

            worker.work()

            # thread = Thread(
            #     target=worker.work,
            #     name=f"Worker#{worker_thread_index}"
            # )
            # self.__thread_manager.register_thread(thread)
            # thread.start()



class TimeCostResult:
    def __init__(self):
        self.start = 0
        self.end = 0
        self.period = 0

    def __str__(self):
        return f"PERIOD {self.period} seconds FROM {self.start} TO {self.end}"



import wave
from typing import List


class WaveWriter:
    def __init__(self, target_wav_file_path: str, number_of_channels: int = 1, sample_width_in_bytes: int = 2,
                 frame_rate: int = 44100):
        # 打开WAV文档
        self.__target_wav_file = wave.open(target_wav_file_path, "wb")
        # 配置声道数、量化位数和取样频率
        self.__number_of_channels = number_of_channels
        self.__target_wav_file.setnchannels(number_of_channels)

        self.__sample_width_in_bytes = sample_width_in_bytes
        self.__target_wav_file.setsampwidth(sample_width_in_bytes)

        self.__frame_rate = frame_rate
        self.__target_wav_file.setframerate(frame_rate)

    def get_number_of_channels(self) -> int:
        """
        声道数
        """
        return self.__number_of_channels

    def get_sample_width_in_bytes(self) -> int:
        """
        量化位数
        """
        return self.__sample_width_in_bytes

    def get_frame_rate(self) -> int:
        """
        取样频率
        """
        return self.__frame_rate

    def write_frames(self, wave_data: List[int]):
        """
        将wav_data转换为二进制数据写入文件
        记录每一个采样点的频率，给定一个int （0-65535）

        WHEN sample_width_in_bytes IS 2
            写入的目标二进制字节串为 0x10 0x27 | 0x9 0x27 | ...
            自 10000=0x2710 | 9993=0x2709 | ... 转换而来
        """

        frames_in_byte = b""
        for frame in wave_data:
            frames_in_byte += frame.to_bytes(2, byteorder="little", signed=True)

        self.__target_wav_file.writeframes(frames_in_byte)

    def __enter__(self):
        pass

    def close(self):
        self.__target_wav_file.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from threading import Lock
from typing import Dict, List

from nehushtan.MessageQueue.NehushtanMessageQueue import NehushtanMessageQueue


class NehushtanMemoryMessageQueue(NehushtanMessageQueue):
    """
    This is a simple implementation for the Message Queue usage in Multi-Threading HTTP Service scenario.
    Since 0.4.12
    """

    def __init__(self):
        super().__init__()

        self.__recycle_lock = Lock()
        self.__lock_map: Dict[str, Lock] = {}
        self.__queue_map: Dict[str, List[str]] = {}

    def __ensure_queue_with_name(self, queue_name: str):
        x = self.__queue_map.get(queue_name)
        if type(x) is not list:
            self.__queue_map[queue_name] = []

    def __lock_queue(self, queue_name: str):
        lock = self.__lock_map.get(queue_name)
        if lock is None:
            self.__lock_map[queue_name] = Lock()
        return self.__lock_map[queue_name].acquire(timeout=1)

    def __unlock_queue(self, queue_name: str):
        lock = self.__lock_map.get(queue_name)
        if lock is not None:
            self.__lock_map[queue_name].release()

    def enqueue(self, item: str, queue_name: str):
        if not self.__lock_queue(queue_name):
            return False

        self.__ensure_queue_with_name(queue_name)
        self.__queue_map[queue_name].append(item)

        self.__unlock_queue(queue_name)

        return True

    def dequeue(self, queue_name: str):
        if not self.__lock_queue(queue_name):
            return False

        x = self.__queue_map.get(queue_name)
        if type(x) is not list:
            return_value = None
        else:
            try:
                return_value = self.__queue_map[queue_name].pop(0)
            except IndexError:
                return None

        self.__unlock_queue(queue_name)
        return return_value

    def stat_for_all(self) -> Dict[str, int]:
        x = {}
        for key in self.__queue_map.keys():
            x[key] = len(self.__queue_map[key])
        return x

    def stat_for_one_queue(self, queue_name: str) -> int:
        x = self.__queue_map.get(queue_name)
        if type(x) is not list:
            return 0
        else:
            return len(x)

    def recycle(self):
        """
        If you want to do a manual recycle, just call this
        """
        self.__recycle_lock.acquire()
        for key in self.__queue_map.keys():
            queue_removed = False
            self.__lock_queue(key)
            if len(self.__queue_map[key]) <= 0:
                del self.__queue_map[key]
                queue_removed = True
            self.__unlock_queue(key)
            if queue_removed and not self.__lock_map[key].locked():
                del self.__lock_map[key]
        self.__recycle_lock.release()



import PySimpleGUI as sg


class SimpleGuiA:
    pass


if __name__ == '__main__':
    sg.theme('DarkAmber')  # Add a touch of color
    # All the stuff inside your window.
    layout = [[sg.Text('Some text on Row 1')],
              [sg.Text('Enter something on Row 2'), sg.InputText()],
              [sg.Button('Ok'), sg.Button('Cancel')]]

    # Create the Window
    window = sg.Window('Window Title', layout)
    # Event Loop to process "events" and get the "values" of the inputs
    while True:
        event, values = window.read()
        if event == sg.WIN_CLOSED or event == 'Cancel':  # if user closes window or clicks cancel
            break
        print('You entered ', values[0])

    window.close()



from nehushtan.logger.legacy.NehushtanFileLogger import NehushtanFileLogger

if __name__ == '__main__':
    log_dir = '/Users/leqee/code/nehushtan/log/x'
    cases = [
        'a',
        'a/b',
        'a/b/c',
        '/a',
        '/a/',
        '/a/b',
        '/a/b/',
        'a/',
        'a/b/',
    ]
    for case in cases:
        NehushtanFileLogger(case, log_dir).info(case)



#  Copyright (c) 2020. Sinri Edogawa
from typing import Iterable, Optional, Union, List

from pymysql.cursors import Cursor

from nehushtan.mysql import constant
from nehushtan.mysql.constant import MYSQL_QUERY_ROW_TYPE_UNKNOWN


class MySQLQueryResult:

    def __init__(self, row_type: str):
        self._sql: str = ''
        self._status: str = constant.MYSQL_QUERY_STATUS_INIT
        self._error: str = 'Not executed yet!'

        self._last_inserted_id: int = -1
        self._affected_rows: int = -1
        self._result_rows: List[Union[tuple, dict]] = []
        self._result_stream: Optional[Cursor] = None

        self._row_type = row_type

    def is_queried(self) -> bool:
        return self._status == constant.MYSQL_QUERY_STATUS_QUERIED

    def is_executed(self) -> bool:
        return self._status == constant.MYSQL_QUERY_STATUS_EXECUTED

    def is_streamed(self) -> bool:
        return self._status == constant.MYSQL_QUERY_STATUS_STREAMED

    def get_last_inserted_id(self) -> int:
        return self._last_inserted_id

    def set_last_inserted_id(self, last_inserted_id: int):
        self._last_inserted_id = last_inserted_id
        return self

    def get_affected_rows(self) -> int:
        return self._affected_rows

    def set_affected_rows(self, affected_rows: int):
        self._affected_rows = affected_rows
        return self

    def get_sql(self) -> str:
        return self._sql

    def set_sql(self, sql: str):
        self._sql = sql
        return self

    def get_status(self) -> str:
        return self._status

    def set_status(self, status):
        self._status = status
        if (
                constant.MYSQL_QUERY_STATUS_QUERIED,
                constant.MYSQL_QUERY_STATUS_STREAMING,
                constant.MYSQL_QUERY_STATUS_STREAMED,
                constant.MYSQL_QUERY_STATUS_EXECUTED,
        ).__contains__(self._status):
            self._error = 'NO ERROR'
        return self

    def get_error(self) -> str:
        return self._error

    def set_error(self, error: str):
        self._error = error
        return self

    def append_result_rows(self, rows: Iterable[Union[tuple, dict]]):
        """

        :param rows: Array (list ot tuple) of result rows, each row would be a dict or tuple
        :return:
        """
        if type(rows) is tuple:
            rows = list(rows)
        self._result_rows += rows
        return self

    def set_stream(self, cursor: Cursor):
        self._result_stream = cursor
        return self

    def get_stream(self) -> Cursor:
        if self._result_stream is None:
            raise IOError("No available stream!")
        return self._result_stream

    def read_next_row(self):
        if self._status != constant.MYSQL_QUERY_STATUS_STREAMING:
            raise IOError("Now no stream to read!")
        row = self.get_stream().fetchone()
        if row is None:
            self._status = constant.MYSQL_QUERY_STATUS_STREAMED
            self._result_stream.close()
            del self._result_stream
            return None
        return row

    # def get_result(self):
    #     """
    #     Removed since 0.3.2
    #     """
    #     warnings.warn('Deprecated since 0.1.12. Use `get_fetched_rows_as_tuple` instead.', DeprecationWarning)
    #     if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
    #         raise Exception('Cannot fetch query result as status is not QUERIED.')
    #     return self._result_rows

    def get_fetched_rows_as_tuple(self):
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise IOError(f'Cannot fetch query result as status is not QUERIED: {self.get_error()}')
        return tuple(self._result_rows)

    def get_column_from_fetched_rows_as_tuple(self, field_key) -> tuple:
        """
        Since 0.3.2
        """
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise IOError(f'Cannot fetch query result as status is not QUERIED: {self.get_error()}')

        column = []
        for row in self._result_rows:
            x = row[field_key]
            column.append(x)

        return tuple(column)

    def get_fetched_first_row_as_tuple(self) -> tuple:
        """
        Since 0.3.2
        """
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise IOError(f'Cannot fetch query result as status is not QUERIED: {self.get_error()}')
        if self._row_type != constant.MYSQL_QUERY_ROW_TYPE_TUPLE:
            raise AssertionError('Row format is not tuple')
        return self._result_rows[0]

    def get_fetched_first_row_as_dict(self) -> dict:
        """
        Since 0.3.2
        """
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise IOError(f'Cannot fetch query result as status is not QUERIED: {self.get_error()}')
        if self._row_type != constant.MYSQL_QUERY_ROW_TYPE_DICT:
            raise AssertionError('Row format is not tuple')
        return self._result_rows[0]

    def get_fetched_first_cell(self, field_key):
        """
        Since 0.3.2
        """
        if self._status != constant.MYSQL_QUERY_STATUS_QUERIED:
            raise IOError(f'Cannot fetch query result as status is not QUERIED: {self.get_error()}')
        return self._result_rows[0][field_key]

    @staticmethod
    def create_error_result(error_message: str):
        return MySQLQueryResult(MYSQL_QUERY_ROW_TYPE_UNKNOWN) \
            .set_status(constant.MYSQL_QUERY_STATUS_ERROR) \
            .set_error(error_message)



import csv
from typing import Optional, IO, Callable, Sequence

from nehushtan.helper.io.CSVReaderOptions import CSVReaderOptions
from nehushtan.helper.io.CSVReaderRowHandleResult import CSVReaderRowHandleResult


class CSVReader:
    """
    https://docs.python.org/zh-cn/3/library/csv.html
    """

    def __init__(self, filename: str, options: Optional[CSVReaderOptions]):
        self.__filename = filename
        self.__csv_file: IO = open(self.__filename, newline='')

        self.__field_names = None

        self.__csv_reader = csv.reader(self.__csv_file, **(options.to_dict()))
        self.__current_index_of_line_to_read = 0

        self.__row_handler: Optional[Callable[[any, int, CSVReaderRowHandleResult], None]] = None
        self.__should_handle_next_row_judge = None
        self.__row_transformer = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__csv_file is not None:
            self.__csv_file.close()

    def total_lines(self) -> int:
        return self.__csv_reader.line_num

    def next_line_index(self) -> int:
        return self.__current_index_of_line_to_read

    def set_field_names(self, field_names: Sequence[str]):
        self.__field_names = field_names
        return self

    def take_first_row_as_field_names(self):
        if self.__current_index_of_line_to_read != 0:
            raise RuntimeError("The first row had been read")
        for row in self.__csv_reader:
            self.set_field_names(row)
            break

    def set_row_handler(self, row_handler: Callable[[any, int, CSVReaderRowHandleResult], None]):
        self.__row_handler = row_handler
        return self

    def _handle_row(self, row, index: int, result: CSVReaderRowHandleResult):
        if self.__row_handler is not None:
            self.__row_handler(row, index, result)

    def set_row_transformer(self, row_transformer: Callable[[any, int, Optional[Sequence[str]]], any]):
        self.__row_transformer = row_transformer
        return self

    def _transform_row(self, row, index: int):
        if self.__row_transformer is None:
            return row
        return self.__row_transformer(row, index, self.__field_names)

    def handle_next_line(self, result: Optional[CSVReaderRowHandleResult] = None):
        """
        row_handler is a callable defined as
            def row_handler(row,index) -> any:
                pass
        """
        for row in self.__csv_reader:
            if result is None:
                result = CSVReaderRowHandleResult()
            transformed_row = self._transform_row(row, self.__current_index_of_line_to_read)
            self._handle_row(transformed_row, self.__current_index_of_line_to_read, result)
            self.__current_index_of_line_to_read += 1
            break

    def handle_next_lines(self, result: Optional[CSVReaderRowHandleResult] = None):
        """
        row_handler is a callable defined as
            def row_handler(row,index) -> bool:
                return should_continue
        """
        for row in self.__csv_reader:
            if result is None:
                result = CSVReaderRowHandleResult()
            transformed_row = self._transform_row(row, self.__current_index_of_line_to_read)
            self._handle_row(transformed_row, self.__current_index_of_line_to_read, result)
            self.__current_index_of_line_to_read += 1
            if not result.get_should_continue():
                break

    @staticmethod
    def basic_row_to_dict_transformer(row, index: int, field_names: Sequence[str]) -> any:
        rest_key = None
        rest_val = None
        rest_list = []

        d = dict()
        for i in range(max(len(row), len(field_names))):
            if i < len(row):
                if i < len(field_names):
                    d[field_names[i]] = row[i]
                else:
                    rest_list.append(row[i])
            else:
                d[field_names[i]] = rest_val

        if len(rest_list) > 0:
            d[rest_key] = rest_list

        return d



from datetime import datetime
from typing import List

from nehushtan.ws.NehushtanWebsocketAgent import NehushtanWebsocketAgent
from nehushtan.ws.NehushtanWebsocketConnectionEntity import NehushtanWebsocketConnectionEntity
from tests.ws.TestWebsocketRegisterAgent import TestWebsocketRegisterAgent


class TestWebsocketAgent(NehushtanWebsocketAgent):
    register_agent: TestWebsocketRegisterAgent

    def register_raw_websocket_connection(self, websocket):
        TestWebsocketAgent.register_agent.register(websocket)

    def unregister_raw_websocket_connection(self, websocket_key: str):
        TestWebsocketAgent.register_agent.unregister(websocket_key)

    def build_response_content(self, websocket_key: str, received_content: str) -> str:
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d %H:%M:%S")

        return f'RESPONSE ON {current_time} TO {websocket_key} FOR {received_content}'

    def get_connections_related_to_this_agent(self) -> List[NehushtanWebsocketConnectionEntity]:
        if not self.get_local_key():
            return []

        return TestWebsocketAgent.register_agent.list_for_server(self.get_local_key())

    def check_content_to_send_to_target_client(self, connection: NehushtanWebsocketConnectionEntity) -> str:
        return f"Broadcast to [{connection.get_key()}]"



import json
import math
import os
import sys
import threading
import time
import traceback
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Optional, List

import psutil

from nehushtan.helper.CommonHelper import CommonHelper


class NehushtanLogLevel(Enum):
    """
    Since 0.5.1.
    """
    CRITICAL = 50
    FATAL = 50
    ERROR = 40
    WARNING = 30
    WARN = 30
    NOTICE = 25
    INFO = 20
    DEBUG = 10
    NOTSET = 0


class NehushtanLoggerAdapter(ABC):
    """
    Since 0.5.1.
    """

    def __init__(self):
        self.__topic = 'default'

    def ensure_serializable_dict(self, d: dict) -> dict:
        """
        Since 0.1.25, add ensure_ascii as False to allow Unicode chars
        """
        s = json.dumps(d, default=lambda inner_x: inner_x.__str__(), ensure_ascii=False)
        return json.loads(s)

    def transform_exception(self, e: BaseException):
        if CommonHelper.is_python_version_at_least(3, 10):
            return traceback.format_exception(e)
        else:
            return traceback.format_exception(
                etype=type(e),
                value=e,
                tb=e.__traceback__
            )

    @abstractmethod
    def write_one_log(self, level: NehushtanLogLevel, contents: dict):
        pass

    def set_topic(self, topic: str):
        self.__topic = topic

    def get_topic(self) -> str:
        return self.__topic


class NehushtanLoggerAdapterWithStdOut(NehushtanLoggerAdapter):
    """
    Since 0.5.1.
    """

    def __init__(self, record_millisecond=False, ):
        super().__init__()
        self.__record_millisecond = record_millisecond

    def _is_record_millisecond(self) -> bool:
        return self.__record_millisecond

    def prepare_log_text(self, level: NehushtanLogLevel, contents: dict):
        time_format_string = "%Y-%m-%d %H:%M:%S"
        if self.__record_millisecond:
            time_format_string += '.%f'
        now = datetime.now().strftime(time_format_string)
        level_label = f'{level.name}'
        pid = os.getpid()

        if CommonHelper.is_python_version_at_least(3, 10):
            thread = threading.current_thread()
        else:
            thread = threading.currentThread()

        clean_contents = self.ensure_serializable_dict(contents)
        message = clean_contents.get('message', '')
        del clean_contents['message']

        line = f'{now} <{self.get_topic()}> [{level_label}] <{pid}:{thread.name}> {message}'
        if len(clean_contents.keys()) > 0:
            extra_json = json.dumps(clean_contents, ensure_ascii=False)
            line += f' | {extra_json}'
        return line

    def write_one_log(self, level: NehushtanLogLevel, contents: dict):
        line = self.prepare_log_text(level, contents)
        print(line)


class NehushtanLoggerAdapterWithFileWriter(NehushtanLoggerAdapterWithStdOut):
    """
    Since 0.5.1.
    """

    def __init__(self,
                 log_dir: str,
                 categorize: bool = True,
                 date_rotate: bool = True,
                 record_millisecond=False,
                 file_encoding='utf-8',
                 ):
        super().__init__(record_millisecond=record_millisecond)

        self.__categorize = categorize

        if log_dir is None:
            raise Exception('NehushtanLoggerAdapterWithFileWriter: log_dir is None')
        self.__log_dir = log_dir

        self.__date_rotate = date_rotate

        # Since 0.4.20 CACHED FILE HANDLER
        self.__keep_file_open = True
        self.__opened_files = {}
        # Since 0.4.20 TARGET FILE ENCODING
        self.__file_encoding = file_encoding

    def set_topic(self, topic: str):
        super().set_topic(topic)
        # This logic is since 0.4.19
        if self.__categorize:
            x = topic.split('/')
            last_title = None
            y = []
            for xx in x:
                if xx:
                    if last_title:
                        y.append(last_title)
                    last_title = xx
            if y and self.__log_dir:
                self.__log_dir = self.__log_dir + '/' + ('/'.join(y))
            if last_title:
                super().set_topic(last_title)

    def __del__(self):
        if len(self.__opened_files.items()) > 0:
            for name, file in self.__opened_files.items():
                file.close()
                # print(name, 'closed')

    def __get_target_file(self):
        if self.__log_dir is None:
            raise Exception('NehushtanLoggerAdapterWithFileWriter: log_dir is None')

        category_dir = self.__log_dir

        # a -> a/a-DATE.log
        # a/b -> a/b-DATE.log
        # a/b/c -> a/b/c-DATE.log

        if self.__categorize:
            category_dir = os.path.join(self.__log_dir, self.get_topic())

        today = ''
        if self.__date_rotate:
            today = time.strftime("%Y%m%d", time.localtime())
            today = f'-{today}'

        target_file = os.path.join(category_dir, f'{self.get_topic()}{today}.log')

        final_dir = os.path.dirname(target_file)
        if not os.path.exists(final_dir):
            os.makedirs(final_dir)

        return target_file

    def __get_target_file_handler(self, target_file_path: str):
        if not target_file_path:
            raise Exception('NehushtanLoggerAdapterWithFileWriter: target_file_path is None')

        if self.__keep_file_open:
            file = self.__opened_files.get(target_file_path)
            if not file:
                file = open(target_file_path, 'a', encoding=self.__file_encoding)
                self.__opened_files[target_file_path] = file
        else:
            file = open(target_file_path, 'a', encoding=self.__file_encoding)

        return file

    def write_raw_line_to_log(self, text: str, end=os.linesep):
        target_file = self.__get_target_file()

        if target_file != '':
            file = self.__get_target_file_handler(target_file)
            file.write(text + end)
            file.flush()

            if not self.__keep_file_open:
                file.close()

        return self

    def write_one_log(self, level: NehushtanLogLevel, contents: dict):
        """
        Since 0.3.7 add `hide_extra`
        """
        line = self.prepare_log_text(level, contents)
        return self.write_raw_line_to_log(line)


class NehushtanLogger:
    """
    Since 0.5.1.
    """

    def __init__(self,
                 topic: str = 'default',
                 adapter: NehushtanLoggerAdapter = NehushtanLoggerAdapterWithStdOut(),
                 log_level: NehushtanLogLevel = NehushtanLogLevel.DEBUG,
                 print_higher_than_this_level: NehushtanLogLevel = NehushtanLogLevel.CRITICAL,
                 default_context: Optional[dict] = None,
                 ):
        self.__topic = topic
        self.__adapter = adapter
        self.__adapter.set_topic(topic)
        self.__log_level = log_level
        self.__default_context = default_context

        # self.print_as_well = print_as_well <-- should use print_higher_than_this_level=NehushtanLogging.NOTSET
        # If all, use NOTSET; if none, use FATAL
        self.__print_higher_than_this_level = print_higher_than_this_level
        if self.__print_higher_than_this_level.value < NehushtanLogLevel.FATAL.value:
            self.__adapter_with_stdout = NehushtanLoggerAdapterWithStdOut()

    def get_adapter(self) -> NehushtanLoggerAdapter:
        return self.__adapter

    def get_traceback_info_from_exception(self, e: BaseException) -> List[str]:
        return self.__adapter.transform_exception(e)

    def write_one_log(self, level: NehushtanLogLevel, contents: dict):
        if self.__default_context is not None:
            context = contents.get('context')
            if context is None:
                contents['context'] = self.__default_context
            else:
                if isinstance(context, dict):
                    c = {}
                    c.update(self.__default_context)
                    c.update(context)
                    contents['context'] = c
                else:
                    pass
        self.get_adapter().write_one_log(level, contents)
        if level.value > self.__print_higher_than_this_level.value:
            self.__adapter_with_stdout.write_one_log(level, contents)

    def _log(self, level: NehushtanLogLevel, message: str, context: Optional[dict] = None):
        if level.value >= self.__log_level.value:
            contents = {'message': message}
            if context is not None:
                contents['context'] = context
            self.write_one_log(level=level, contents=contents)

    def debug(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.DEBUG, message=message, context=extra)

    def info(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.INFO, message=message, context=extra)

    def notice(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.NOTICE, message=message, context=extra)

    def warning(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.WARNING, message=message, context=extra)

    def error(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.ERROR, message=message, context=extra)

    def critical(self, message: str, extra: Optional[dict] = None):
        self._log(level=NehushtanLogLevel.CRITICAL, message=message, context=extra)

    def exception(self, message: str, e: BaseException):
        level = NehushtanLogLevel.ERROR
        traces = self.get_adapter().transform_exception(e)
        if level.value >= self.__log_level.value:
            contents = {
                'message': message,
                'exception': {
                    'class': e.__class__.__name__,
                    'message': e.__str__(),
                    'trace': traces,
                }
            }
            self.write_one_log(level=level, contents=contents)

    def log_progress(
            self,
            title: str,
            done_task_count: int,
            total_task_count: int = 100,
            progress_bar_length: int = 20,
            desc: str = None,
            level: NehushtanLogLevel = NehushtanLogLevel.NOTICE
    ):
        percent = 100.0 * done_task_count / total_task_count
        bar = ''
        done_bar_chars = math.floor(progress_bar_length * percent / 100.0)
        for i in range(done_bar_chars):
            bar += '='
        for j in range(progress_bar_length - done_bar_chars):
            bar += '-'

        content = f'{title}: {percent:2.2f}% ({done_task_count}/{total_task_count}) [{bar}]'
        if desc is not None:
            content += f' {desc}'

        self._log(level=level, message=content, )

    def log_current_memory_usage_of_process(self, pid: int = None, level: NehushtanLogLevel = NehushtanLogLevel.INFO):
        """
        Filed `pid` would use os.getpid() for None.
        """
        memory_usage = psutil.Process(pid=pid).memory_info()
        self._log(
            level,
            'Current Memory Usage Snapshot (in MB)',
            {'rss': memory_usage.rss / 1024.0 / 1024.0, 'vms': memory_usage.vms / 1024.0 / 1024.0}
        )

    def log_current_memory_usage_of_object(self, target_name: str, target,
                                           level: NehushtanLogLevel = NehushtanLogLevel.INFO):

        memory_usage = sys.getsizeof(target)
        self._log(
            level,
            f'Current Memory Usage used by {target_name} (in MB)',
            {'size': memory_usage / 1024.0 / 1024.0}
        )



class NehushtanHTTPConstant:
    """
    Since 0.4.0
    """

    METHOD_HEAD = "HEAD"
    METHOD_GET = "GET"
    METHOD_POST = "POST"
    METHOD_PUT = "PUT"
    METHOD_DELETE = "DELETE"
    METHOD_OPTIONS = "OPTIONS"
    METHOD_PATCH = "PATCH"

    HEADER_CONTENT_TYPE = 'Content-Type'

    HEADER_CONTENT_TYPE_VALUE_TEXT = 'text/plain'
    HEADER_CONTENT_TYPE_VALUE_HTML = 'text/html'
    HEADER_CONTENT_TYPE_VALUE_JSON = 'application/json'

    REPLY_CODE_OK = 'OK'
    REPLY_CODE_FAIL = 'FAIL'



from typing import Iterable


class MySQLJoinOnCondition:
    def __init__(self, left_field_expression: str, right_field_expression: str):
        self.__left_field_expression = left_field_expression
        self.__right_field_expression = right_field_expression

    def __str__(self):
        return self.organize_to_sql()

    def organize_to_sql(self):
        return self.__left_field_expression + " = " + self.__right_field_expression

    @staticmethod
    def make_field_equal(left: Iterable[str], right: Iterable[str]):
        """
        Provide `left` and `right` each as
            1) a string for table name (alias);
        or  2) a tuple of two strings, for schema and table names (aliases).
        """
        left_field_expression = '`' + '`.`'.join(left) + '`'
        right_field_expression = '`' + '`.`'.join(right) + '`'

        return MySQLJoinOnCondition(
            left_field_expression=left_field_expression,
            right_field_expression=right_field_expression,
        )


import socket
import time
from threading import Thread
from typing import Optional

from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLoggerAdapterWithFileWriter, NehushtanLogLevel
from nehushtan.socket.SocketHandleThreadManager import SocketHandlerThreadManager


class NehushtanWormholeWorker:
    """
    Since 0.4.28
    """

    def __init__(self,
                 socket_name: str,
                 upstream_socket: socket.socket, upstream_address,
                 downstream_host: str, downstream_port: int,
                 log_dir: str = None,
                 thread_manager: SocketHandlerThreadManager = None
                 ):
        self.__socket_name = socket_name

        self.__upstream_socket = upstream_socket
        self.__upstream_address = upstream_address

        self.__downstream_socket: Optional[socket.socket] = None
        self.__downstream_host = downstream_host
        self.__downstream_port = downstream_port

        # self.__upstream_socket.settimeout(1)
        # self.__downstream_socket.settimeout(1)

        # self.__data_queue_from_upstream = Queue()
        # self.__data_queue_from_downstream = Queue()

        # socket_name = f"{time.time()}#{CommonHelper.generate_random_uuid_hex()}"
        if log_dir is None:
            self.__logger = NehushtanLogger(topic=f"WormholeSocket/{socket_name}")
        else:
            self.__logger = NehushtanLogger(topic=f"WormholeSocket/{socket_name}",
                                            adapter=NehushtanLoggerAdapterWithFileWriter(
                                                log_dir=log_dir,
                                            ))
        self.__logger.notice(f"From {upstream_address}")

        self.__thread_manager = thread_manager

    def set_logger(self, logger: NehushtanLogger):
        self.__logger = logger
        return self

    def work(self):
        address_entries = socket.getaddrinfo(
            self.__downstream_host,
            self.__downstream_port,
            socket.AF_UNSPEC,
            socket.SOCK_STREAM
        )

        prepare_error = None

        for address_entry in address_entries:
            af, socktype, proto, canonname, sa = address_entry
            try:
                self.__downstream_socket = socket.socket(af, socktype, proto)
                self.__downstream_socket.connect(sa)
            except OSError as msg:
                prepare_error = msg
                if self.__downstream_socket is not None:
                    self.__downstream_socket.close()
                self.__downstream_socket = None
                continue
            break

        if self.__downstream_socket is None:
            raise prepare_error

        self.__logger.info(f"Downstream Socket Connected to {self.__downstream_socket.getpeername()}")

        thread_reading_upstream = Thread(
            target=self.thread_body_reading_upstream,
            name=f"{self.__socket_name}::upstream_monitor",
            daemon=True
        )
        thread_reading_downstream = Thread(
            target=self.thread_body_reading_downstream,
            name=f"{self.__socket_name}::downstream_monitor",
            daemon=True
        )

        thread_reading_upstream.start()
        thread_reading_downstream.start()

        self.__logger.info("Two reading threads started")

        self.__thread_manager.register_thread(thread_reading_upstream)
        self.__thread_manager.register_thread(thread_reading_downstream)

        self.__logger.debug("registered two threads and finish work method")

    @staticmethod
    def read_all_available_from_socket(stream_socket: socket.socket, buffer_size: int, logger: NehushtanLogger):
        total_buffer = b""
        while True:
            try:
                buffer = stream_socket.recv(buffer_size, socket.MSG_DONTWAIT)
                if len(buffer) <= 0:
                    break
                logger.debug(f"read_all_available_from_socket one piece of {len(buffer)} bytes")
                total_buffer += buffer
                time.sleep(0.1)
            except BlockingIOError:
                if len(total_buffer) > 0:
                    break
                else:
                    raise
        return total_buffer

    def thread_body_reading_upstream(self):
        while True:
            # self.__logger.debug("thread_body_reading_upstream routine start")
            try:
                buffer = self.read_all_available_from_socket(self.__upstream_socket, 1024, self.__logger)
                # buffer = self.__upstream_socket.recv(1024, socket.MSG_DONTWAIT)
                if len(buffer) <= 0:
                    break
                self.__logger.info(f"Read {len(buffer)} bytes from upstream")
            except BlockingIOError:
                # self.__logger.debug(f"thread_body_reading_upstream met BlockingIOError, sleep")
                time.sleep(1)
                continue
            except OSError:
                self.__logger.error("Upstream Dead")
                break

            self.__logger.write_one_log(NehushtanLogLevel.DEBUG, {'buffer': f"{buffer}"})
            try:
                write_byte_count = self.__downstream_socket.send(buffer)
                if write_byte_count > 0:
                    self.__logger.info(f"Wrote {write_byte_count} bytes to downstream")
                else:
                    self.__logger.error(f"Failed to write into downstream")
                    break
            except OSError:
                self.__logger.error("Upstream Dead")
                break

        self.__logger.notice("CLOSE SOCKETS")
        self.__downstream_socket.close()
        self.__upstream_socket.close()

    def thread_body_reading_downstream(self):
        while True:
            # self.__logger.debug("thread_body_reading_downstream routine start")
            try:
                buffer = self.read_all_available_from_socket(self.__downstream_socket, 1024, self.__logger)
                # buffer = self.__downstream_socket.recv(1024, socket.MSG_DONTWAIT)

                if len(buffer) <= 0:
                    break

                self.__logger.info(f"Read {len(buffer)} bytes from downstream")

            except BlockingIOError:
                # self.__logger.debug(f"thread_body_reading_downstream met BlockingIOError, sleep")
                time.sleep(1)
                continue
            except OSError:
                self.__logger.error("Downstream Dead")
                break

            self.__logger.write_one_log(NehushtanLogLevel.DEBUG, {'buffer': f"{buffer}"})
            try:
                write_byte_count = self.__upstream_socket.send(buffer)
                if write_byte_count > 0:
                    self.__logger.info(f"Write {write_byte_count} bytes to upstream")
                else:
                    self.__logger.error(f"Failed to write into upstream")
                    break
            except OSError:
                self.__logger.error("Upstream Dead")
                break

        self.__logger.notice("CLOSE SOCKETS")
        self.__downstream_socket.close()
        self.__upstream_socket.close()



from typing import List

from nehushtan.httpd.NehushtanHTTPRoute import NehushtanHTTPRoute
from nehushtan.httpd.exceptions.NehushtanNoRouteMatchedError import NehushtanNoRouteMatchedError
from nehushtan.logger.NehushtanLogger import NehushtanLogger


class NehushtanHTTPRouter:
    """
    Since 0.4.0
    """

    def __init__(self, logger: NehushtanLogger = None):
        self.__route_list: List[NehushtanHTTPRoute] = []
        self.__logger = logger

    def register_route(self, route: NehushtanHTTPRoute):
        self.__route_list.append(route)

        if self.__logger is not None:
            self.__logger.debug(f'NehushtanHTTPRouter::register_route {route}')

        return self

    def check_request_for_route(self, method: str, path: str) -> NehushtanHTTPRoute:
        for route in self.__route_list:
            matched = route.match_request(method, path)
            if self.__logger is not None:
                self.__logger.debug(f'NehushtanHTTPRouter::check_request_for_route',
                                    {'matched': matched, 'method': method, 'path': path, 'route': route})
            if matched:
                return route

        raise NehushtanNoRouteMatchedError(
            http_error_message=f'Request [{method} {path}] Matched No Route.',
            http_code=404
        )



import shutil

from nehushtan.logger.NehushtanLogger import NehushtanLogger, NehushtanLoggerAdapterWithFileWriter
from tests.queue.plan2.Test2NehushtanQueue import Test2NehushtanQueue
from tests.queue.plan2.Test2NehushtanQueueDelegate import Test2NehushtanQueueDelegate
from tests.queue.plan2.Test2NehushtanQueueTaskDelegate import Test2NehushtanQueueTaskDelegate

if __name__ == '__main__':
    shutil.rmtree('path/to/nehushtan/log/queue-plan2')

    config_dictionary = {}
    logger = NehushtanLogger(topic='loop',
                             adapter=NehushtanLoggerAdapterWithFileWriter(
                                 log_dir='path/to/nehushtan/log/queue-plan2'))
    delegate = Test2NehushtanQueueDelegate(config_dictionary, logger)
    task_delegate = Test2NehushtanQueueTaskDelegate(config_dictionary, logger)

    Test2NehushtanQueue(delegate, task_delegate).loop()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='nehushtan',
    version='0.5.13',
    packages=find_packages(
        exclude=['tests', 'tests.*']
    ),
    url='https://sinri.github.io/nehushtan/',
    license='MIT',
    author='Sinri Edogawa',
    author_email='e.joshua.s.e@gmail.com',
    description='A toolkit for projects in Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.8',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'psutil~=6.0.0',
    ]
)



import json
from typing import Literal

from nehushtan.score.ParsedLine import ParsedLine


class ParsedLineAsLyric(ParsedLine):
    def __init__(self, content: str, prefix=None, content_align_type: Literal["<", "=", ">"] = ">"):
        super().__init__()
        self.__prefix = prefix
        self.__content = content
        self.__content_align_type = content_align_type

    def get_prefix(self):
        return self.__prefix

    def get_content(self):
        return self.__content

    def get_lyric_char(self, index: int):
        return self.__content[index]

    def get_content_align_type(self):
        return self.__content_align_type

    def __str__(self):
        x = (self.__prefix, self.__content)
        return "ParsedLineAsLyric: " + json.dumps(x)



from PySimpleGUI import PySimpleGUI

from nehushtan.gui.simple.SimpleWindow import SimpleWindow


class SimpleGuiB(SimpleWindow):

    def __init__(self):
        super().__init__('SimpleGuiB')

        self.__graph = PySimpleGUI.Graph(
            graph_bottom_left=(0, 0),
            graph_top_right=(200, 100),
            background_color='orange',
            canvas_size=(200, 100)
        )

    def get_current_layout(self) -> list[list]:
        # # All the stuff inside your window.

        layout = [
            [
                # PySimpleGUI.Text('Some text on Row 1')
                self.__graph
            ],
            [PySimpleGUI.Text('y='), PySimpleGUI.InputText(default_text="x", key="formula")],
            [
                PySimpleGUI.Text('Ready', key='msg'),
                PySimpleGUI.Button('Draw'),
                # PySimpleGUI.Button('Cancel')
            ]
        ]
        return layout

    def should_continue_loop_after_event(self, event, values) -> bool:
        print('should_continue_loop_after_event', event, values)
        if event == PySimpleGUI.WIN_CLOSED:
            return False
        # if event == 'Cancel':
        #     print("cancelled")
        #     return False
        if event == 'Draw':
            self.get_window().set_title(values[0])

            try:
                x = 0
                while x <= 200:
                    s: str = values['formula']
                    s.replace("x", f'{x}')
                    y = eval(s)
                    self.__graph.draw_point((x, y))
                    x += 0.1
            except Exception as e:
                print(e)
                self.get_window()['msg'].update(e.__str__())

            return True


if __name__ == '__main__':
    sw = SimpleGuiB()
    sw.start_loop()



from http.server import HTTPServer, BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Callable


class NehushtanHTTPService:
    """
    Since 0.4.0
    """

    def __init__(self):
        self.__server = None

    def get_server(self) -> HTTPServer:
        return self.__server

    def set_server(self, server: HTTPServer):
        self.__server = server
        return self

    def run(self, poll_interval=0.5):
        self.get_server().serve_forever(poll_interval=poll_interval)

    @staticmethod
    def run_with_selector_server(
            handler_class: Callable[..., BaseHTTPRequestHandler],
            listen_port: int,
            listen_host=''
    ):
        """
        USE HTTPServer
        This class builds on the TCPServer class by storing the server address
        as instance variables named server_name and server_port.
        The server is accessible by the handler, typically through the handler’s server instance variable.
        """
        server_address = (listen_host, listen_port)
        httpd = HTTPServer(server_address, handler_class)
        httpd.serve_forever()

    @staticmethod
    def run_with_threading_server(
            handler_class: Callable[..., BaseHTTPRequestHandler],
            listen_port: int,
            listen_host=''
    ):
        """
        USE ThreadingHTTPServer
        This class is identical to HTTPServer but uses threads to handle requests by using the ThreadingMixIn.
        This is useful to handle web browsers pre-opening sockets, on which HTTPServer would wait indefinitely.
        New in version 3.7.
        """
        server_address = (listen_host, listen_port)
        httpd = ThreadingHTTPServer(server_address, handler_class)
        httpd.serve_forever()



from typing import List, Tuple

from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLSelectionMixin import MySQLSelectionMixin
from nehushtan.mysql.MySQLTableSelection import MySQLTableSelection


class MySQLUnionSelection(MySQLSelectionMixin):
    """
    Since 0.3.6
    """

    def __init__(self):
        super().__init__()
        self.sub_selection_list: List[Tuple[MySQLTableSelection, str]] = []

    def get_mysql_kit(self) -> MySQLKit:
        """
        Just use the MySQLKit instance of the first sub selection
        """
        if len(self.sub_selection_list) <= 0:
            raise AssertionError('Sub Selection List should not be empty.')
        return self.sub_selection_list[0][0].get_mysql_kit()

    def add_sub_selection(self, sub_selection: MySQLTableSelection, use_all: bool = False):
        union_type = 'UNION'
        if use_all:
            union_type += ' ALL'
        self.sub_selection_list.append((sub_selection, union_type))
        return self

    def generate_sql(self) -> str:
        sqls = ''
        for i in range(len(self.sub_selection_list)):
            if i > 0:
                sqls += ' ' + self.sub_selection_list[i][1] + ' '
            sqls += self.sub_selection_list[i][0].generate_sql()
        return sqls



import re
from typing import Sequence, Tuple, Union

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.httpd.NehushtanHTTPRoute import NehushtanHTTPRoute


class NehushtanHTTPRouteWithRestFul(NehushtanHTTPRoute):
    """
    Since 0.4.0
    """

    def __init__(
            self,
            path_template: str,
            process_chain_base_package: str,
            filter_list: Sequence[Union[type, str]] = None,
            method_options: Sequence[str] = None
    ):
        """
        `path_template` is expected as '/PRE../..FIX',
            would be mapped to URL '/PRE../..FIX/PACKAGE...S/CLASS/METHOD'
            or '/PRE../..FIX/PACKAGE...S/CLASS/METHOD/ARGS[0]/ARGS[1]/...'
        `process_chain_base_package` should be a package namespace name, e.g. 'a.b.c',
            (thus, the target class-method pair for '/d/e/f' would be (`a.b.c.d.e`,`f`)
        """
        super().__init__()
        self.path_template = path_template
        self.process_chain_base_package = process_chain_base_package
        self.sub_class_path = ''
        self.class_method_name = ''
        self.filter_list = filter_list if filter_list is not None else []
        self.method_options = method_options

    def match_request(self, method: str, path: str) -> bool:
        if self.method_options is not None:
            if not self.method_options.__contains__(method):
                return False

        # print(self.path_template)

        regex = r'^(' + self.path_template + ')/(.+)$'
        result = re.search(regex, path)
        if result is None:
            return False
        # print(regex)
        # print(result)
        components = result.group(2).split('/')
        # print(components)

        class_full_path = self.process_chain_base_package
        for i in range(len(components)):
            component = components[i]
            class_full_path += '.' + component
            try:
                CommonHelper.class_with_class_path(class_full_path)
            except ModuleNotFoundError:
                continue
            if len(components) - 1 == i:
                # do not have enough path components
                return False
            self.sub_class_path = ".".join(components[:i + 1])
            self.class_method_name = components[i + 1]
            self.matched_arguments = components[i + 2:]
            # print({
            #     'self.sub_class_path':self.sub_class_path,
            #     'self.class_method_name':self.class_method_name,
            #     'self.matched_arguments':self.matched_arguments,
            # })
            return True

        # anyway not match
        return False

        # self.class_method_name = result.group(2)
        #
        # if len(result.groups()) > 2:
        #     x = result.group(2)[1:].split('/')
        #     self.matched_arguments = x
        # return True

    def get_filter_list(self) -> Sequence[Union[type, str]]:
        return self.filter_list

    def get_controller_target(self) -> Tuple[Union[type, str], str]:
        return self.process_chain_base_package + "." + self.sub_class_path, self.class_method_name

    def __str__(self):
        method = "|".join(self.method_options) if self.method_options is not None else 'ANY'
        return f'NehushtanHTTPRouteWithRestFul [{method}] {self.path_template}'



#  Copyright (c) 2020. Sinri Edogawa

import os
import smtplib
from email.header import Header
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import parseaddr, formataddr
import warnings
from typing import Iterable


class SMTPAgent:
    # __smtp_server: str
    # __smtp_port: int
    #
    # __password: str
    #
    # _sender: str
    # _receivers: Iterable[str]
    # _subject: str
    # _content: str
    # _attachments: Iterable[str]

    def __init__(self, sender: str, password: str, smtp_server: str, smtp_port: int):
        self.__smtp_server = smtp_server
        self.__smtp_port = smtp_port
        self._sender = sender
        self.__password = password

        self._receivers = []
        self._subject = ''
        self._content = ''
        self._attachments = []

    def set_receivers(self, receivers: Iterable[str]):
        warnings.warn("Deprecated since 0.1.11. Use `reset_receivers` or `add-` methods instead!", DeprecationWarning)
        return self.reset_receivers(receivers)

    def reset_receivers(self, receivers: Iterable[str]):
        self._receivers = []
        self.add_receivers(receivers)
        return self

    def add_receivers(self, receivers: Iterable[str]):
        for receiver in receivers:
            self.add_receiver(receiver)
        return self

    def add_receiver(self, receiver: str):
        self._receivers.append(receiver)
        return self

    def set_subject(self, subject: str):
        self._subject = subject
        return self

    def set_content(self, content: str):
        self._content = content
        return self

    def set_attachments(self, attachments: Iterable[str]):
        warnings.warn("Deprecated since 0.1.11. Use `reset_attachments` or `add-` methods instead!", DeprecationWarning)
        return self.reset_attachments(attachments)

    def reset_attachments(self, attachments: Iterable[str]):
        self._attachments = []
        self.add_attachments(attachments)
        return self

    def add_attachment(self, attachment_file_path: str):
        self._attachments.append(attachment_file_path)
        return self

    def add_attachments(self, attachment_file_path_array: Iterable[str]):
        for attachment_file_path in attachment_file_path_array:
            self.add_attachment(attachment_file_path)
        return self

    @staticmethod
    def _format_address(s):
        name, addr = parseaddr(s)
        return formataddr(
            (
                Header(name, 'utf-8').encode(),
                addr
            )
        )

    def send(self):
        message_html = MIMEMultipart()
        message_html['From'] = self._format_address(self._sender)

        to_list = []
        for receiver in self._receivers:
            to_list.append(self._format_address(receiver))

        message_html['To'] = ','.join(to_list)

        message_html['Subject'] = Header(self._subject, 'gbk')
        message_html.attach(MIMEText(self._content, 'html', 'utf-8'))

        for attachment in self._attachments:
            file_name = os.path.basename(attachment)
            part = MIMEApplication(open(attachment, 'rb').read())
            part.add_header('Content-Disposition', 'attachment', filename=file_name)  # 给附件重命名,一般和原文件名一样,改错了可能无法打开.
            message_html.attach(part)

        server = smtplib.SMTP_SSL(self.__smtp_server, self.__smtp_port)
        server.login(self._sender, self.__password)
        server.sendmail(self._sender, self._receivers, message_html.as_string())
        server.quit()

        return self



from nehushtan.score.NotationTextParser import NotationTextParser
from nehushtan.score.ScoreDrawer import ScoreDrawer
from nehushtan.score.ScoreDrawerOptions import ScoreDrawerOptions

if __name__ == '__main__':
    file = "/debug/score/00"
    parser = NotationTextParser(file=file)
    parser.parse()
    # parser.debug()
    font1 = "/Users/leqee/code/Verbum/fonts/Arial Unicode.ttf"
    font2 = "/Users/leqee/code/Verbum/fonts/FangZhengKaiTi-GBK/FangZhengKaiTi-GBK-1.ttf"

    options = ScoreDrawerOptions(font1, 50)
    options.set_lyric_font_path(font1)
    options.set_lyric_font_size(40)
    options.set_score_font_size(40)
    options.set_title_left_component_font_size(30)
    options.set_title_right_component_font_size(30)
    options.set_title_middle_component_font_size(70)

    options.set_total_cells_in_one_line(32)

    drawer = ScoreDrawer(parser.get_parsed_score_lines(), options)
    # drawer.show("test4")
    drawer.save("/Users/leqee/code/nehushtan/debug/score/book.png")



from nehushtan.socket.NehushtanTCPSocketClient import NehushtanTCPSocketClient


class EchoClient(NehushtanTCPSocketClient):

    def handle_client_connection(self):
        connection = self.get_connection()

        while True:
            x = input("Say: ")
            if not x:
                break

            sent = connection.send(x.encode())
            print(f'[DEBUG] handle_client_connection sent bytes: {sent}')
            received = connection.recv(1024)
            print(f'Heard from {connection.getpeername()}: {received}')

        self.close_connection()


if __name__ == '__main__':
    EchoClient(4444).listen()



from typing import Literal

from PIL import Image, ImageFont
from PIL import ImageDraw


class Paint:
    """
    https://pillow.readthedocs.io/en/stable/reference/ImageDraw.html
    """

    def __init__(self, width: int = 400, height: int = 600, mode=None, background_color=None):
        """
        mode: https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes
        """
        if not mode:
            mode = 'RGB'

        self.__image = Image.new(mode, (width, height))

        self.__draw = ImageDraw.Draw(self.__image)

        if background_color is not None:
            self.__draw.rectangle(
                ((0, 0),
                 (width, height)),
                fill=background_color
            )

    def get_image(self):
        return self.__image

    def get_draw_of_image(self):
        return self.__draw

    def arc(self, xy, start, end, fill=None, width=0):
        """
        Draws an arc (a portion of a circle outline) between the start and end angles, inside the given bounding box.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of [(x0, y0), (x1, y1)] or [x0, y0, x1, y1], where x1 >= x0 and y1 >= y0.
            start – Starting angle, in degrees. Angles are measured from 3 o’clock, increasing clockwise.
            end – Ending angle, in degrees.
            fill – Color to use for the arc.
            width – The line width, in pixels. New in version 5.3.0.
        """
        self.__draw.arc(xy, start, end, fill, width)

    def chord(self, xy, start, end, fill=None, outline=None, width=1):
        """
        Same as arc(), but connects the end points with a straight line.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of [(x0, y0), (x1, y1)] or [x0, y0, x1, y1], where x1 >= x0 and y1 >= y0.
            outline – Color to use for the outline.
            fill – Color to use for the fill.
            width – The line width, in pixels. New in version 5.3.0.
        """
        self.__draw.chord(xy, start, end, fill, outline, width)

    def pieslice(self, xy, start, end, fill=None, outline=None, width=1):
        """
        Same as arc, but also draws straight lines between the end points and the center of the bounding box.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of [(x0, y0), (x1, y1)] or [x0, y0, x1, y1], where x1 >= x0 and y1 >= y0.
            start – Starting angle, in degrees. Angles are measured from 3 o’clock, increasing clockwise.
            end – Ending angle, in degrees.
            fill – Color to use for the fill.
            outline – Color to use for the outline.
            width – The line width, in pixels. New in version 5.3.0.
        """
        self.__draw.pieslice(xy, start, end, fill, outline, width)

    def ellipse(self, xy, fill=None, outline=None, width=1):
        """
        Draws an ellipse inside the given bounding box.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of either [(x0, y0), (x1, y1)] or [x0, y0, x1, y1], where x1 >= x0 and y1 >= y0.
            outline – Color to use for the outline.
            fill – Color to use for the fill.
            width – The line width, in pixels. New in version 5.3.0.
        """
        self.__draw.ellipse(xy, fill, outline, width)

    def line(self, xy, fill=None, width=0, joint=None):
        """
        Draws a line between the coordinates in the xy list.

        PARAMETERS:
            xy – Sequence of either 2-tuples like [(x, y), (x, y), ...] or numeric values like [x, y, x, y, ...].
            fill – Color to use for the line.
            width – The line width, in pixels.
                New in version 1.1.5.
                Note This option was broken until version 1.1.6.
            joint – Joint type between a sequence of lines. It can be "curve", for rounded edges, or None.
                New in version 5.3.0.
        """
        self.__draw.line(xy, fill, width, joint)

    def point(self, xy, fill=None):
        """
        Draws points (individual pixels) at the given coordinates.

        PARAMETERS:
            xy – Sequence of either 2-tuples like [(x, y), (x, y), ...] or numeric values like [x, y, x, y, ...].
            fill – Color to use for the point.

        """
        self.__draw.point(xy, fill)

    def polygon(self, xy, fill=None, outline=None, width=1):
        """
        Draws a polygon.

        The polygon outline consists of straight lines between the given coordinates, plus a straight line between the last and the first coordinate.

        PARAMETERS:
            xy – Sequence of either 2-tuples like [(x, y), (x, y), ...] or numeric values like [x, y, x, y, ...].
            fill – Color to use for the fill.
            outline – Color to use for the outline.
            width – The line width, in pixels.
        """
        self.__draw.polygon(xy.xy, fill, outline, width)

    def regular_polygon(self, bounding_circle, n_sides, rotation=0, fill=None, outline=None):
        """
        Draws a regular polygon inscribed in bounding_circle, with n_sides, and rotation of rotation degrees.

        PARAMETERS:
            bounding_circle – The bounding circle is a tuple defined by a point and radius. (e.g. bounding_circle=(x, y, r) or ((x, y), r)). The polygon is inscribed in this circle.
            n_sides – Number of sides (e.g. n_sides=3 for a triangle, 6 for a hexagon).
            rotation – Apply an arbitrary rotation to the polygon (e.g. rotation=90, applies a 90 degree rotation).
            fill – Color to use for the fill.
            outline – Color to use for the outline.
        """
        self.__draw.regular_polygon(bounding_circle, n_sides, rotation, fill, outline)

    def rectangle(self, xy, fill=None, outline=None, width=1):
        """
        Draws a rectangle.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of either [(x0, y0), (x1, y1)] or [x0, y0, x1, y1]. The second point is just outside the drawn rectangle.
            outline – Color to use for the outline.
            fill – Color to use for the fill.
            width – The line width, in pixels. New in version 5.3.0.
        """
        self.__draw.rectangle(xy, fill, outline, width)

    def rounded_rectangle(self, xy, radius=0, fill=None, outline=None, width=1):
        """
        Draws a rounded rectangle.

        PARAMETERS:
            xy – Two points to define the bounding box. Sequence of either [(x0, y0), (x1, y1)] or [x0, y0, x1, y1]. The second point is just outside the drawn rectangle.
            radius – Radius of the corners.
            outline – Color to use for the outline.
            fill – Color to use for the fill.
            width – The line width, in pixels.

        New in version 8.2.0.
        """
        self.__draw.rounded_rectangle(xy, radius, fill, outline, width)

    def text(self, xy, text, fill=None, font=None, anchor=None, spacing=4,
             align: Literal["left", "center", "right"] = 'left', direction=None, features=None,
             language=None, stroke_width=0, stroke_fill=None, embedded_color=False):
        """
        Draws the string at the given position.

        PARAMETERS:
            xy – The anchor coordinates of the text.
            text – String to be drawn. If it contains any newline characters, the text is passed on to multiline_text().
            fill – Color to use for the text.
            font – An ImageFont instance.
            anchor – The text anchor alignment. Determines the relative location of the anchor to the text. The default alignment is top left. See Text anchors for valid values. This parameter is ignored for non-TrueType fonts.
                Note This parameter was present in earlier versions of Pillow, but implemented only in version 8.0.0.
            spacing – If the text is passed on to multiline_text(), the number of pixels between lines.
            align – If the text is passed on to multiline_text(), "left", "center" or "right". Determines the relative alignment of lines. Use the anchor parameter to specify the alignment to xy.
            direction – Direction of the text. It can be "rtl" (right to left), "ltr" (left to right) or "ttb" (top to bottom). Requires libraqm.
                New in version 4.2.0.
            features – A list of OpenType font features to be used during text layout. This is usually used to turn on optional font features that are not enabled by default, for example "dlig" or "ss01", but can be also used to turn off default font features, for example "-liga" to disable ligatures or "-kern" to disable kerning. To get all supported features, see OpenType docs. Requires libraqm.
                New in version 4.2.0.
            language –  Language of the text. Different languages may use different glyph shapes or ligatures. This parameter tells the font which language the text is in, and to apply the correct substitutions as appropriate, if available. It should be a BCP 47 language code. Requires libraqm.
                New in version 6.0.0.
            stroke_width – The width of the text stroke.
                New in version 6.2.0.
            stroke_fill – Color to use for the text stroke. If not given, will default to the fill parameter.
                New in version 6.2.0.
            embedded_color – Whether to use font embedded color glyphs (COLR, CBDT, SBIX).
                New in version 8.0.0.
        """
        self.__draw.text(xy, text, fill, font, anchor, spacing, align, direction, features, language, stroke_width,
                         stroke_fill, embedded_color)

    def multiline_text(self, xy, text, fill=None, font=None, anchor=None, spacing=4,
                       align: Literal["left", "center", "right"] = 'left', direction=None, features=None, language=None,
                       stroke_width=0, stroke_fill=None, embedded_color=False):
        """
        Draws the string at the given position.

        PARAMETERS:
            xy – The anchor coordinates of the text.
            text – String to be drawn.
            fill – Color to use for the text.
            font – An ImageFont instance.
            anchor – The text anchor alignment. Determines the relative location of the anchor to the text. The default alignment is top left. See Text anchors for valid values. This parameter is ignored for non-TrueType fonts.
                Note This parameter was present in earlier versions of Pillow, but implemented only in version 8.0.0.
            spacing – The number of pixels between lines.
            align – "left", "center" or "right". Determines the relative alignment of lines. Use the anchor parameter to specify the alignment to xy.
            direction – Direction of the text. It can be "rtl" (right to left), "ltr" (left to right) or "ttb" (top to bottom). Requires libraqm.
                New in version 4.2.0.
            features – A list of OpenType font features to be used during text layout. This is usually used to turn on optional font features that are not enabled by default, for example "dlig" or "ss01", but can be also used to turn off default font features, for example "-liga" to disable ligatures or "-kern" to disable kerning. To get all supported features, see OpenType docs. Requires libraqm.
                New in version 4.2.0.
            language – Language of the text. Different languages may use different glyph shapes or ligatures. This parameter tells the font which language the text is in, and to apply the correct substitutions as appropriate, if available. It should be a BCP 47 language code. Requires libraqm.
                New in version 6.0.0.
            stroke_width – The width of the text stroke.
                New in version 6.2.0.
            stroke_fill – Color to use for the text stroke. If not given, will default to the fill parameter.
                New in version 6.2.0.
            embedded_color – Whether to use font embedded color glyphs (COLR, CBDT, SBIX).
                New in version 8.0.0.
        """
        self.__draw.multiline_text(xy, text, fill, font, anchor, spacing, align, direction, features, language,
                                   stroke_width, stroke_fill, embedded_color)

    def textlength(self, text, font=None, direction=None, features=None, language=None, embedded_color=False):
        """
        Returns length (in pixels with 1/64 precision) of given text when rendered in font with provided direction, features, and language.
        This is the amount by which following text should be offset. Text bounding box may extend past the length in some fonts, e.g. when using italics or accents.
        The result is returned as a float; it is a whole number if using basic layout.
        Note that the sum of two lengths may not equal the length of a concatenated string due to kerning. If you need to adjust for kerning, include the following character and subtract its length.

        New in version 8.0.0.

        PARAMETERS:
            text – Text to be measured. May not contain any newline characters.
            font – An ImageFont instance.
            direction – Direction of the text. It can be "rtl" (right to left), "ltr" (left to right) or "ttb" (top to bottom). Requires libraqm.
            features – A list of OpenType font features to be used during text layout. This is usually used to turn on optional font features that are not enabled by default, for example "dlig" or "ss01", but can be also used to turn off default font features, for example "-liga" to disable ligatures or "-kern" to disable kerning. To get all supported features, see OpenType docs. Requires libraqm.
            language – Language of the text. Different languages may use different glyph shapes or ligatures. This parameter tells the font which language the text is in, and to apply the correct substitutions as appropriate, if available. It should be a BCP 47 language code. Requires libraqm.
            embedded_color – Whether to use font embedded color glyphs (COLR, CBDT, SBIX).
        """
        return self.__draw.textlength(text, font, direction, features, language, embedded_color)

    def textbbox(self, xy, text, font=None, anchor=None, spacing=4, align: Literal["left", "center", "right"] = 'left',
                 direction=None, features=None, language=None, stroke_width=0, embedded_color=False):
        """
        Returns bounding box (in pixels) of given text relative to given anchor when rendered in font with provided direction, features, and language. Only supported for TrueType fonts.

        Use textlength() to get the offset of following text with 1/64 pixel precision. The bounding box includes extra margins for some fonts, e.g. italics or accents.

        New in version 8.0.0.

        PARAMETERS:
            xy – The anchor coordinates of the text.
            text – Text to be measured. If it contains any newline characters, the text is passed on to multiline_textbbox().
            font – A FreeTypeFont instance.
            anchor – The text anchor alignment. Determines the relative location of the anchor to the text. The default alignment is top left. See Text anchors for valid values. This parameter is ignored for non-TrueType fonts.
            spacing – If the text is passed on to multiline_textbbox(), the number of pixels between lines.
            align – If the text is passed on to multiline_textbbox(), "left", "center" or "right". Determines the relative alignment of lines. Use the anchor parameter to specify the alignment to xy.
            direction – Direction of the text. It can be "rtl" (right to left), "ltr" (left to right) or "ttb" (top to bottom). Requires libraqm.
            features – A list of OpenType font features to be used during text layout. This is usually used to turn on optional font features that are not enabled by default, for example "dlig" or "ss01", but can be also used to turn off default font features, for example "-liga" to disable ligatures or "-kern" to disable kerning. To get all supported features, see OpenType docs. Requires libraqm.
            language – Language of the text. Different languages may use different glyph shapes or ligatures. This parameter tells the font which language the text is in, and to apply the correct substitutions as appropriate, if available. It should be a BCP 47 language code. Requires libraqm.
            stroke_width – The width of the text stroke.
            embedded_color – Whether to use font embedded color glyphs (COLR, CBDT, SBIX).
        """
        return self.__draw.textbbox(xy, text, font, anchor, spacing, align, direction, features, language, stroke_width,
                                    embedded_color)

    def multiline_textbbox(self, xy, text, font=None, anchor=None, spacing=4,
                           align: Literal["left", "center", "right"] = 'left', direction=None, features=None,
                           language=None, stroke_width=0, embedded_color=False):
        """
        Returns bounding box (in pixels) of given text relative to given anchor when rendered in font with provided direction, features, and language. Only supported for TrueType fonts.

        Use textlength() to get the offset of following text with 1/64 pixel precision. The bounding box includes extra margins for some fonts, e.g. italics or accents.

        New in version 8.0.0.

        PARAMETERS:
            xy – The anchor coordinates of the text.
            text – Text to be measured.
            font – A FreeTypeFont instance.
            anchor – The text anchor alignment. Determines the relative location of the anchor to the text. The default alignment is top left. See Text anchors for valid values. This parameter is ignored for non-TrueType fonts.
            spacing – The number of pixels between lines.
            align – "left", "center" or "right". Determines the relative alignment of lines. Use the anchor parameter to specify the alignment to xy.
            direction – Direction of the text. It can be "rtl" (right to left), "ltr" (left to right) or "ttb" (top to bottom). Requires libraqm.
            features – A list of OpenType font features to be used during text layout. This is usually used to turn on optional font features that are not enabled by default, for example "dlig" or "ss01", but can be also used to turn off default font features, for example "-liga" to disable ligatures or "-kern" to disable kerning. To get all supported features, see OpenType docs. Requires libraqm.
            language – Language of the text. Different languages may use different glyph shapes or ligatures. This parameter tells the font which language the text is in, and to apply the correct substitutions as appropriate, if available. It should be a BCP 47 language code. Requires libraqm.
            stroke_width – The width of the text stroke.
            embedded_color – Whether to use font embedded color glyphs (COLR, CBDT, SBIX).
        """
        return self.__draw.multiline_textbbox(xy, text, font, anchor, spacing, align, direction, features, language,
                                              stroke_width, embedded_color)

    def save(self, fp, the_format=None, **params):
        """
        Saves this image under the given filename. If no format is specified, the format to use is determined from the filename extension, if possible.

        Keyword options can be used to provide additional instructions to the writer. If a writer doesn’t recognise an option, it is silently ignored. The available options are described in the image format documentation for each writer.

        You can use a file object instead of a filename. In this case, you must always specify the format. The file object must implement the seek, tell, and write methods, and be opened in binary mode.

        PARAMETERS:
            fp – A filename (string), pathlib.Path object or file object.
            format – Optional format override. If omitted, the format to use is determined from the filename extension. If a file object was used instead of a filename, this parameter should always be used.
            params – Extra parameters to the image writer.

        RETURNS: None

        RAISES:
            ValueError – If the output format could not be determined from the file name. Use the format option to solve this.
            OSError – If the file could not be written. The file may have been created, and may contain partial data.
        """
        self.__image.save(fp, the_format, **params)

    def show(self, title: str = None):
        """
        Displays this image. This method is mainly intended for debugging purposes.

        This method calls PIL.ImageShow.show() internally. You can use PIL.ImageShow.register() to override its default behaviour.

        The image is first saved to a temporary file. By default, it will be in PNG format.

        On Unix, the image is then opened using the display, eog or xv utility, depending on which one can be found.

        On macOS, the image is opened with the native Preview application.

        On Windows, the image is opened with the standard PNG display utility.

        PARAMETERS:
            title – Optional title to use for the image window, where possible.
        """
        self.__image.show(title)

    def close(self):
        self.__image.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def load_ttf_font(ttf_font_file_path: str, size: int):
        # use a truetype font
        return ImageFont.truetype(ttf_font_file_path, size)
